/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

public class SocketChannelNatives {

	/**
	 * Connects the socket associated with the file descriptor {@code fd} to the remote host address {@code addr} and {@code port}
	 * with a specified {@code timeout} value.
	 * <p>A timeout of zero is interpreted as an infinite timeout. The connection will then block until established or an error occurs.
	 * <p>The timeout detection is done by the caller of this method so the implementation does not need to detect a timeout. In any
	 * case, this method must block for more than the specified timeout (if timeout is different from 0).
	 * 
	 * @param fd the socket file descriptor
	 * @param addr the remote host address buffer
	 * @param length the remote host address size
	 * @param port the remote host port
	 * @param timeout the timeout value in milliseconds
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 1 if the connection succeeded; 0 if the file descriptor is in non-blocking mode and the connection did not immediately succeed
	 * or negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int connect(int fd, byte[] addr, int length, int port, int timeout, boolean retry);

	/**
	 * Returns the local port to which the socket associated with the file descriptor {@code fd} is bound.
	 * 
	 * @param fd the socket file descriptor
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the local port to which the socket associated with the file descriptor {@code fd} is bound;
	 * -1 if the socket is not bound or a negative error code.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getLocalPort(int fd, boolean retry);

	/**
	 * Returns the remote host port to which the socket associated with the file descriptor {@code fd} is connected.
	 * 
	 * @param fd the socket file descriptor
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the remote port to which the socket associated with the file descriptor {@code fd} is connected;
	 * or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getPeerPort(int fd, boolean retry);

	/**
	 * Gets the local address to which the socket associated with the file descriptor {@code fd} is bound.
	 * When this methods returns, the {@code addr} buffer is filled with the local address.
	 * 
	 * @param fd the socket file descriptor
	 * @param addr the buffer into which the local address will be stored
	 * @param length the buffer length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the local address size in bytes, 0 if the socket is not bound or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getLocalAddress(int fd, byte[] addr, int length, boolean retry);

	/**
	 * Gets the remote host address to which the socket associated with the file descriptor {@code fd} is connected.
	 * When this methods returns, the {@code addr} buffer is filled with the remote host address.
	 * 
	 * @param fd the socket file descriptor
	 * @param addr the buffer into which the remote host address will be stored
	 * @param length the buffer length
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the remote host address size in bytes, 0 if the socket is not bound or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getPeerAddress(int fd, byte[] addr, int length, boolean retry);

	/**
	 * Creates a new socket and returns its associated file descriptor.
	 * 
	 * @param stream set to true for streaming sockets; false for datagrams.
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the file descriptor of the new socket or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int socket(boolean stream, boolean retry);


	/**
	 * Creates a new server socket, returning the native file descriptor.
	 * 
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the file descriptor of the new server socket or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int serverSocket(boolean retry);

	/**
	 * Creates a new multicast socket, returning the native file descriptor.
	 * 
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the file descriptor of the new server socket or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int multicastSocket(boolean retry);
}
