/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl;

import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateEncodingException;

import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContextSpi;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;

import com.is2t.support.net.ssl.x509.X509KeyManagerImpl;
import com.is2t.support.net.ssl.x509.X509TrustManagerImpl;

public abstract class AbstractSSLContext extends SSLContextSpi {


	private KeyManager keyManager;
	private TrustManager trustManager;

	// underlying SSL Client Context
	private int clientContextID;
	// underlying SSL Server Context
	private int serverContextID;

	private int clientSocketCount;
	private int serverSocketCount;

	private boolean isInitialized = false;

	/**
	 * SSL/TLS protocol versions
	 *
	 */
	public enum ProtocolVersion {

		SSLv3, // SSL v3

		TLSv1, // TLS v1.0

		TLSV1_1, // TLS v1.1

		TLSV1_2, // TLS v1.2

		DTLSv1, // DTLS v1.0

		DTLSv1_2, // DTLS v1.2

		TLSV1_3, // TLS v1.3
	}


	@Override
	protected void engineInit(KeyManager[] kms, TrustManager[] tms, SecureRandom sr) throws KeyManagementException {

		if (this.isInitialized) {
			throw new KeyManagementException("context already initialized");
		}

		// choose the appropriate KeyManager
		this.keyManager = chooseKeyManager(kms);

		this.trustManager = null;
		if (tms == null) {
			try {
				// get the default TrustManagers
				TrustManagerFactory tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
				tmf.init(null);
				tms = tmf.getTrustManagers();
			} catch (Exception e) {
				// ignore
			}
		}

		// choose the first trust manager that is an implementation of
		// X509TrustManagerImpl
		this.trustManager = chooseTrustManager(tms);

		this.isInitialized = true;

	}

	/*
	 * Create and initialize the underlying context according to the mode
	 * (server or client)
	 */
	private void initUnderlyingContext(boolean isServer) throws SSLException {
		int contextID;
		if (isServer) {
			contextID = this.serverContextID = SSLContextSupport.createContext(getProtocol(), false);
		} else {
			contextID = this.clientContextID = SSLContextSupport.createContext(getProtocol(), true);
		}
		try {
			// initialize the context with the keystore credentials
			((X509KeyManagerImpl) this.keyManager).initContextKeyStore(contextID);
			// initialize the context truststore with the trusted certs
			((X509TrustManagerImpl) this.trustManager).initContextTrustStore(contextID);

		} catch (KeyManagementException | CertificateEncodingException e) {
			// init failed, free the context
			SSLContextSupport.closeContext(contextID);
			throw new SSLException(e);
		}
	}

	/*
	 * Called when a underlying socket using this context is going to be
	 * created. This callback checks if the context need to be initialized
	 */
	void beforeCreateSocketCallback(boolean socketIsServer) throws SSLException {
		if (socketIsServer) {
			if (this.serverSocketCount == 0) {
				// no created socket (in server mode) associated with this context
				// create and initialize the underlying server context
				initUnderlyingContext(true);
			}
		} else {
			if (this.clientSocketCount == 0) {
				// no created socket (in client mode) associated with this context
				// initialize the underlying client context
				initUnderlyingContext(false);
			}
		}
	}

	/*
	 * Called when a underlying socket using this context is created or it
	 * creation failed. The parameter isServer is <code>true</code> if the
	 * socket is created in server mode or <code>false</code> if the socket is
	 * created in client mode.
	 */
	void afterCreateSocketCallback(boolean isServer, boolean socketCreationFailed) {
		if (isServer) {
			if (!socketCreationFailed) {
				// +1 socket (in server mode) created; increment the server
				// counter
				this.serverSocketCount++;
			} else if (this.serverSocketCount == 0) {
				// The creation of the underlying server socket failed but a
				// server context has been created in the previous call of
				// {@link #beforeCreateSocketCallback(boolean)}.
				// Just free this context
				try {
					SSLContextSupport.closeContext(this.serverContextID);
				} catch (Exception e) {
					// ignore
				}
			}
			// else: there are existing sockets (in server mode) associated to
			// this context => nothing to do
		} else {
			if (!socketCreationFailed) {
				// +1 socket (in client mode) created; increment the client
				// counter
				this.clientSocketCount++;
			} else if (this.clientSocketCount == 0) {
				// The creation of the underlying client socket failed but a
				// client context has been created in the previous call of
				// {@link #beforeCreateSocketCallback(boolean)}.
				// Just free this context
				try {
					SSLContextSupport.closeContext(this.clientContextID);
				} catch (Exception e) {
					// ignore
				}
			}
			// else: there are existing sockets (in client mode) associated to
			// this context => nothing to do
		}
	}

	/*
	 * Called when a created socket is closed
	 */
	void closeSocketCallback(boolean isServer) {
		try {
			if (isServer) {
				if (this.serverSocketCount > 0) {
					this.serverSocketCount--;
				}
				if (this.serverSocketCount == 0) {
					// all sockets (in server mode) created from this context
					// are closed; free the context
					SSLContextSupport.closeContext(this.serverContextID);
				}
			} else {
				if (this.clientSocketCount > 0) {
					this.clientSocketCount--;
				}
				if (this.clientSocketCount == 0) {
					SSLContextSupport.closeContext(clientContextID);
				}
			}
		} catch (SSLException e) {
			// ignore
		}
	}

	@Override
	protected SSLSocketFactory engineGetSocketFactory() {
		checkInit();
		return new SSLSocketFactoryImpl(this);
	}

	@Override
	protected SSLServerSocketFactory engineGetServerSocketFactory() {
		checkInit();
		return new SSLServerSocketFactoryImpl(this);
	}

	/**
	 * Returns the SSLContext protocol version.
	 *
	 * @return returns the SSLContext protocol version.
	 */
	protected abstract ProtocolVersion getProtocol();


	int getClientContextID(boolean isServer) {
		return isServer ? this.serverContextID : this.clientContextID;
	}


	private void checkInit() {
		if (!this.isInitialized) {
			throw new IllegalStateException("SSL Context is not initialized");
		}
	}

	private KeyManager chooseKeyManager(KeyManager[] kms) throws KeyManagementException {
		for (int i = 0; kms != null && i < kms.length; i++) {
			KeyManager current = kms[i];
			if (current instanceof X509KeyManagerImpl) {
				return current;
			}
		}
		try {
			// Dummy KeyManager instance
			return new X509KeyManagerImpl(null, null);
		} catch (UnrecoverableKeyException | KeyStoreException | NoSuchAlgorithmException e) {
			throw new KeyManagementException(e);
		}
	}

	private TrustManager chooseTrustManager(TrustManager[] tms) {
		for (int i = 0; tms != null && i < tms.length; i++) {
			TrustManager current = tms[i];
			if (current instanceof X509TrustManagerImpl) {
				return current;
			}
		}
		// Dummy TrustManager instance;
		return new X509TrustManagerImpl(null);
	}

}