/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2016-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net.ssl;

import java.security.KeyManagementException;
import java.security.cert.CertificateEncodingException;

import javax.net.ssl.SSLException;

import com.is2t.support.net.ssl.AbstractSSLContext.ProtocolVersion;
import com.is2t.support.net.ssl.natives.SSLContextSupportNatives;
import com.is2t.support.security.x509.X509CertImpl;
import com.is2t.support.security.x509.X509KeyImpl;

import ej.error.Message;
import ej.sni.SNI;

public class SSLContextSupport {

	static {
		SSLSocketSupport.initialize();
	}

	public static int createContext(ProtocolVersion protocol, boolean useClientMode) throws SSLException {
		int protocolID = protocol.ordinal();
		int ret = SSLContextSupportNatives.createContext(protocolID, useClientMode, false);

		// retry native returned with result
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLContextSupportNatives.createContext(protocolID, useClientMode, true);
		}

		if (ret == SSLErrorConstants.J_CREATE_SSL_CONTEXT_ERROR) {
			throw new SSLException("Could not create the underlying SSLContext");
		}
		return ret;
	}

	public static void closeContext(int ctxID) throws SSLException {
		int ret = SSLContextSupportNatives.closeContext(ctxID, false);

		// retry native returned with result
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLContextSupportNatives.closeContext(ctxID, true);
		}

		if (ret < 0) {
			throw new SSLException("Could not release the underlying SSLContext");
		}
	}

	public static void addTrustedCert(int contextID, X509CertImpl cert)
			throws KeyManagementException, CertificateEncodingException {
		synchronized (SSLResources.CERT_BUFFER) {
			byte[] certBuffer = SSLResources.CERT_BUFFER;
			int len = cert.getEncoded().length;

			try {
				System.arraycopy(cert.getEncoded(), 0, certBuffer, 0, len);

			} catch (IndexOutOfBoundsException e) {
				throw new KeyManagementException(
						Message.at(SSLErrorMessages.getInstance(), SSLErrorConstants.J_BIG_CERT_OR_KEY_SIZE_ERROR));
			}

			int format = cert.getEncodingFormat().ordinal();
			int ret = SSLContextSupportNatives.addTrustedCert(contextID, certBuffer, 0, len, format, false);

			// retry native returned with result
			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SSLContextSupportNatives.addTrustedCert(contextID, certBuffer, 0, len, format, true);
			}

			if (ret < 0) {
				throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
			}
		}

	}

	public static void clearTrustStore(int contextID) throws KeyManagementException {
		int ret = SSLContextSupportNatives.clearTrustStore(contextID, false);

		// retry native returned with result
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLContextSupportNatives.clearTrustStore(contextID, true);
		}

		if (ret < 0) {
			throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
		}
	}

	public static void clearKeyStore(int contextID) throws KeyManagementException {
		int ret = SSLContextSupportNatives.clearKeyStore(contextID, false);

		// retry native returned without result
		while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = SSLContextSupportNatives.clearKeyStore(contextID, true);
		}

		if (ret < 0) {
			throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
		}

	}

	public static void useCredentials(int contextID, X509KeyImpl key, String password,
			X509CertImpl[] certs) throws KeyManagementException, CertificateEncodingException {

		if (key != null && certs != null && certs.length > 0) {

			// on first, set the certificate associated with the private key
			// this cert is at index 0
			setCertificate(contextID, certs[0]);

			// set the private key
			setPrivateKey(contextID, key, password);

			// compute the total size of the rest of the certificate (the
			// issuers chain certificates)
			int chainCertsTotalSize = 0;
			int nbChainCerts = certs.length - 1;
			for (int i = 1; i <= nbChainCerts; i++) {
				chainCertsTotalSize += certs[i].getEncoded().length;
			}

			// init the chain buffer
			int chainBufferSize = initChainBuffer(contextID, nbChainCerts, chainCertsTotalSize);

			// set the certificate chain if any
			for (int i = 1; i <= nbChainCerts; i++) {
				addChainCertificate(contextID, certs[i], chainBufferSize);
			}


		}
	}

	private static int initChainBuffer(int contextID, int nbChainCerts, int chainCertsTotalSize)
			throws KeyManagementException {
		int chainBufferSize = SSLContextSupportNatives.initChainBuffer(contextID, nbChainCerts, chainCertsTotalSize,
				false);
		// retry - native returned without result
		while (chainBufferSize == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			chainBufferSize = SSLContextSupportNatives.initChainBuffer(contextID, nbChainCerts, chainCertsTotalSize,
					true);
		}
		if (chainBufferSize < 0) {
			throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), chainBufferSize));
		}
		return chainBufferSize;
	}

	private static void setCertificate(int contextID, X509CertImpl cert)
			throws KeyManagementException, CertificateEncodingException {
		synchronized (SSLResources.CERT_BUFFER) {
			byte[] buffer = SSLResources.CERT_BUFFER;
			int len = cert.getEncoded().length;
			try {
				System.arraycopy(cert.getEncoded(), 0, buffer, 0, len);

			} catch (IndexOutOfBoundsException e) {
				throw new KeyManagementException(
						Message.at(SSLErrorMessages.getInstance(), SSLErrorConstants.J_BIG_CERT_OR_KEY_SIZE_ERROR));
			}
			int format = cert.getEncodingFormat().ordinal();
			int ret = SSLContextSupportNatives.setCertificate(contextID, buffer, 0, len, format, false);
			// retry - native returned without result
			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SSLContextSupportNatives.setCertificate(contextID, buffer, 0, len, format, true);
			}

			if (ret < 0) {
				throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
			}
		}
	}

	private static void setPrivateKey(int contextID, X509KeyImpl key, String password) throws KeyManagementException {
		synchronized (SSLResources.CERT_BUFFER) {
			byte[] buffer = SSLResources.CERT_BUFFER;
			int len = key.getKeyLength();
			try {
				System.arraycopy(key.getKey(), 0, buffer, 0, len);

			} catch (IndexOutOfBoundsException e) {
				throw new KeyManagementException(
						Message.at(SSLErrorMessages.getInstance(), SSLErrorConstants.J_BIG_CERT_OR_KEY_SIZE_ERROR));
			}
			synchronized (SSLResources.KEY_PASSWORD_BUFFER) {
				byte[] passphrase = SSLResources.KEY_PASSWORD_BUFFER;
				int passphraseLen = 0;
				if (password != null) {
					passphraseLen = password.length();
					try {
						SNI.toCString(password, passphrase);

					} catch (IndexOutOfBoundsException e) {
						throw new KeyManagementException(
								Message.at(SSLErrorMessages.getInstance(), SSLErrorConstants.J_BIG_KEY_PASSWORD_SIZE_ERROR));
					}
				}

				int ret = SSLContextSupportNatives.setPrivateKey(contextID, buffer, 0, len, passphrase, 0,
						passphraseLen, false);
				// retry - native returned without result
				while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
					ret = SSLContextSupportNatives.setPrivateKey(contextID, buffer, 0, len, passphrase, 0,
							passphraseLen, true);
				}

				if (ret < 0) {
					throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
				}
			}
		}
	}

	private static void addChainCertificate(int contextID, X509CertImpl cert, int chainBufferSize)
			throws KeyManagementException, CertificateEncodingException {
		synchronized (SSLResources.CERT_BUFFER) {
			byte[] buffer = SSLResources.CERT_BUFFER;
			int len = cert.getEncoded().length;
			try {

				System.arraycopy(cert.getEncoded(), 0, buffer, 0, len);
			} catch (IndexOutOfBoundsException e) {
				throw new KeyManagementException(
						Message.at(SSLErrorMessages.getInstance(), SSLErrorConstants.J_BIG_CERT_OR_KEY_SIZE_ERROR));
			}
			int format = cert.getEncodingFormat().ordinal();
			int ret = SSLContextSupportNatives.addChainCertificate(contextID, buffer, 0, len, format, chainBufferSize,
					false);
			// retry - native returned without result
			while (ret == SSLErrorConstants.J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				ret = SSLContextSupportNatives.addChainCertificate(contextID, buffer, 0, len, format, chainBufferSize,
						true);
			}

			if (ret < 0) {
				throw new KeyManagementException(Message.at(SSLErrorMessages.getInstance(), ret));
			}
		}
	}


}
