/*
 * Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.digest;

import java.security.DigestException;
import java.security.MessageDigestSpi;
import java.security.ProviderException;

/**
 * Common base message digest implementation for the Sun provider.
 * It implements all the JCA methods as suitable for a Java message digest
 * implementation of an algorithm based on a compression function (as all
 * commonly used algorithms are). The individual digest subclasses only need to
 * implement the following methods:
 *
 *  . abstract void implCompress(byte[] b, int ofs);
 *  . abstract void implDigest(byte[] out, int ofs);
 *  . abstract void implReset();
 *
 * See the inline documentation for details.
 *
 * Implementation copied from <code>sun.security.provider.DigestBase</code>.
 *
 * @since   1.5
 * @author  Andreas Sterbenz
 */
abstract public class AbstractMessageDigestSpi extends MessageDigestSpi implements Cloneable {

	// one element byte array, temporary storage for update(byte)
	private byte[] oneByte;

	// length of the message digest in bytes
	private final int digestLength;

	// number of bytes processed so far. subclasses should not modify
	// this value.
	// also used as a flag to indicate reset status
	// -1: need to call engineReset() before next call to update()
	//  0: is already reset
	long bytesProcessed;

	/**
	 * Main constructor.
	 */
	AbstractMessageDigestSpi(int digestLength) {
		super();
		this.digestLength = digestLength;
	}

	// return digest length. See JCA doc.
	@Override
	protected final int engineGetDigestLength() {
		return digestLength;
	}

	// single byte update. See JCA doc.
	@Override
	protected final void engineUpdate(byte b) {
		if (oneByte == null) {
			oneByte = new byte[1];
		}
		oneByte[0] = b;
		engineUpdate(oneByte, 0, 1);
	}

	// array update. See JCA doc.
	@Override
	protected final void engineUpdate(byte[] b, int ofs, int len) {
		if (len == 0) {
			return;
		}
		if ((ofs < 0) || (len < 0) || (ofs > b.length - len)) {
			throw new ArrayIndexOutOfBoundsException();
		}
		if (bytesProcessed < 0) {
			engineReset();
		}

		implUpdate(b, ofs, len);
	}

	// reset this object. See JCA doc.
	@Override
	protected void engineReset() {
		if (bytesProcessed == 0) {
			// already reset, ignore
			return;
		}
		implReset();
		bytesProcessed = 0;
	}

	// return the digest. See JCA doc.
	@Override
	protected final byte[] engineDigest() {
		byte[] b = new byte[digestLength];
		try {
			engineDigest(b, 0, b.length);
		} catch (DigestException e) {
			throw (ProviderException)
			new ProviderException("Internal error").initCause(e);
		}
		return b;
	}

	// return the digest in the specified array. See JCA doc.
	@Override
	protected final int engineDigest(byte[] out, int ofs, int len)
			throws DigestException {
		if (len < digestLength) {
			throw new DigestException("Length must be at least "
					+ digestLength + " for this digest");
		}
		if ((ofs < 0) || (len < 0) || (ofs > out.length - len)) {
			throw new DigestException("Buffer too short to store digest");
		}
		if (bytesProcessed < 0) {
			engineReset();
		}
		implDigest(out, ofs, digestLength);
		bytesProcessed = -1;
		return digestLength;
	}

	/**
	 * TODO comment
	 */
	abstract void implUpdate(byte[] b, int ofs, int len);

	/**
	 * Return the digest. Subclasses do not need to reset() themselves,
	 * DigestBase calls implReset() when necessary.
	 */
	abstract void implDigest(byte[] out, int ofs, int len);

	/**
	 * Reset subclass specific state to their initial values. DigestBase
	 * calls this method when necessary.
	 */
	abstract void implReset();

}
