/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */
package demo.mqtt;

import java.io.IOException;

import org.eclipse.paho110.client.mqttv311.IMqttDeliveryToken;
import org.eclipse.paho110.client.mqttv311.MqttCallback;
import org.eclipse.paho110.client.mqttv311.MqttClient;
import org.eclipse.paho110.client.mqttv311.MqttConnectOptions;
import org.eclipse.paho110.client.mqttv311.MqttException;
import org.eclipse.paho110.client.mqttv311.MqttMessage;
import org.eclipse.paho110.client.mqttv311.internal.ClientComms;
import org.eclipse.paho110.client.mqttv311.persist.MemoryPersistence;

import com.cwm.net.SSLMicroNetworkModule;

import ej.ecom.mobile.ATConnection;
import ej.ecom.mobile.PDPSession;
import ej.gnss.GnssManager;

/*
 * THALES Developer Zone article is here, with patched "Library" sources for J2ME 3.2:
 * https://developer.gemalto.com/showcase/paho-project-110-july-2016-mqtt-311
 *
 * Library - Paho 1.1.0 client library source code for Java SE7 originally from here:
 * https://repo.eclipse.org/content/repositories/paho-releases/
 *
 * HelloWorld - Paho 1.1.0 Getting Started Java SE sample code originally from here:
 * https://github.com/eclipse/paho.mqtt.java (see bottom of that page for code)
 * https://www.eclipse.org/paho/clients/java/
 *
 * Conceptualising the connection to a broker and seeing the results:
 * https://www.hivemq.com/blog/mqtt-essentials-part2-publish-subscribe
 *
 * Practical real world examples:
 * 1. Eclipse broker is here:
 *    tcp://iot.eclipse.org:1883    - connect to iot.eclipse.org on port 1883 using plain TCP
 *    tls://iot.eclipse.org:8883    - connect to iot.eclipse.org on port 8883 using secure TLS
 *                                  - ssl:// and tcps:// are also synonyms for tls://
 *
 * 2. HiveMQ provides a free to test Broker (as of 11.2018)
 *    Along with a nice Dashboard and Web Sockets interface to see whats going on:
 *    tcp://broker.hivemq.com:1883   - broker in plain text (no security)
 *    http://www.mqtt-dashboard.com/ - dashboard web interface
 *    http://www.hivemq.com/demos/websocket-client/ - web interface to view messages
 *
 */

public class Main {

	private static String broker      = null;   // choose plain text or secure broker
	private static String brokerCert = null;
	private static String brokerUser  = null;   // sign is as broker user
	private static String brokerPass  = null;   // sign in with broker password
	private static String setPIN = null;
	private static String setAPN = null; // choose APN for SIM
	private static String setUser = null; // choose Username for SIM's APN
	private static String setPass = null; // choose Password for SIM's APN
	private static String pubTopic = null; // choose MQTT publication topic
	private boolean skipGnss = false;

	protected void startApp() {
		int pubQos          = 0;    // really just "fire and forget"; QoS = 1 is better though can create duplicate publishes
		String pubContent   = "Hello World! from Thales-IoT-Module";
		String clientId     = "Thales-IoT-Module";
		String location = null;

		boolean useSecure = false;
		String brokerNormal = "tcp://broker.hivemq.com:1883";   // normal TCP sockets connect - see http://www.mqtt-dashboard.com/
		String brokerSecure = "ssl://mqtt.eclipse.org:8883"; // secure TLS /SSL sockets connect

		System.out.println("");
		System.out.println("[Main]: PAHO Version = '" + org.eclipse.paho.client.mqttv3.internal.ClientComms.VERSION + "' if left unfactored.");
		System.out.println("");

		if ((ClientComms.VERSION != null) && (ClientComms.VERSION.indexOf("SJM") == -1)) {
			System.out.println("[Main]: Warning PAHO Version = " + ClientComms.VERSION);
			System.out.println("[Main]: Please use our refactored PAHO library available here:");
			System.out.println("[Main]: https://developer.gemalto.com/showcase/paho-project-110-july-2016-mqtt-311");
		}

		readSettingsFromJADFile();          // Here we read the APN and MQTT broker details from the .JAD file

		if (broker == null) {
			broker = useSecure ? brokerNormal // send data to the broker in plain text, with no TLS encryption
					: brokerSecure; // send data to the broker using TLS encryption
		}

		if (setAPN == null) {
			System.out.println("[Main]: Error no SIM APN defined, need 'setAPN:' in .JAD file");
		}
		if (broker == null) {
			System.out.println("[Main]: Error no MQTT broker defined, need 'broker:' in .JAD file");
		}
		if (setAPN == null || broker == null) {
			return;
		}

		if (brokerCert != null && !brokerCert.isEmpty()) {
			SSLMicroNetworkModule.setCertificatePath(brokerCert);
		}

		ATConnection atc = ATConnection.getInstance();

		clientId = readIMEI(atc, clientId); // We need a unique clientId if possible

		try {
			// TODO check if already unlocked
			System.out.println("[Main]: Unlocking SIM");
			unlockSim(atc);
		} catch (IOException e) {
			System.out.println("[Main]: Sim unlock fails: " + e.toString());
		}

		if (!skipGnss) {
			// First fix may take longer
			location = getPosition(atc, 10);
		}

		while (true) {
			publishInfo(pubQos, pubContent, clientId, location, atc);
			if (!skipGnss) {
				location = getPosition(atc, 1);
			}
		}
	}

	private String getPosition(ATConnection atc, int retry) {
		GnssManager gnssManager = GnssManager.getInstance();
		String position = null;
		long time = 0;

		try {
			System.out.println("[Main]: Starting the GNSS engine.");
			setAirplaneMode(atc);
			startGnssEngine(atc);
			gnssManager.switchOn();

			System.out.println("[Main]: Locating... (May take a while)");
			for (int i = 1; i <= retry && time == 0; i++) {
				if (i > 1) {
					System.out.println("[Main]: Trying again... (" + i + "/" + retry + ")");
				}
				gnssManager.readPosition();
				time = gnssManager.getTime();
			}
			if (time != 0) {
				System.out.println("[Main]: Location found");
				float latitude = gnssManager.getLatitude();
				float longitude = gnssManager.getLongitude();
				position = "{ \"PCB latitude\": " + floatToString(latitude) + ", \"PCB longitude\": "
						+ floatToString(longitude) + " }";
			} else {
				System.out.println("[Main]: Error detecting location: timeout.");
			}
		} catch (IOException e) {
			System.out.println("[Main]: Error detecting location: " + e.getMessage());
		} finally {
			System.out.println("[Main]: Stoping the GNSS engine.");
			try {
				clearAtBuffer(atc);
				gnssManager.switchOff();
			} catch (IOException e) {
				System.out.println("[Main]: Error switching GPS off: " + e.getMessage());
			}
			try {
				stopGnssEngine(atc);
			} catch (IOException e) {
				System.out.println("[Main]: Error stoping GPS engine: " + e.getMessage());
			}
			try {
				setNormalMode(atc);
			} catch (IOException e) {
				System.out.println("[Main]: Error setting mode to normal: " + e.getMessage());
			}
		}

		return position;
	}

	private void setNormalMode(ATConnection atc) throws IOException {
		String response;
		sendCommand(atc, "AT+CFUN=1,0\r");
		response = readResponse(atc);
		if (!response.contains("OK")) {
			throw new IOException("AT^CFUN command failed: " + response);
		}
	}

	private void stopGnssEngine(ATConnection atc) throws IOException {
		String response;
		sendCommand(atc, "AT^SGPSC=\"Engine\",0\r");
		response = readResponse(atc);
		if (!response.contains("OK") || !response.contains("^SGPSC: \"Engine\",\"0\"")) {
			throw new IOException("AT^SGPSC command failed: " + response);
		}
	}

	private void startGnssEngine(ATConnection atc) throws IOException {
		String response;
		sendCommand(atc, "AT^SGPSC=\"Engine\",3\r");
		response = readResponse(atc);
		if (!response.contains("OK") || !response.contains("^SGPSC: \"Engine\",\"3\"")) {
			throw new IOException("AT^SGPSC command failed: " + response);
		}
	}

	private void setAirplaneMode(ATConnection atc) throws IOException {
		String response;
		sendCommand(atc, "AT+CFUN=4,0\r");
		response = readResponse(atc);
		if (!response.contains("OK")) {
			throw new IOException("AT^CFUN command failed: " + response);
		}
	}

	private void publishInfo(int pubQos, String pubContent, String clientId, String location, ATConnection atc) {
		try {
			System.out.println("[Main]: Attaching network (may take a while)");
			while (!attachNetwork(atc)) {
				System.out.println("[Main]: Trying again...");
			}

			// Define the PDP context
			System.out.println("[Main]: Configuring the PDP context");
			configurePDP(atc);

			// Start a PDP session
			System.out.println("[Main]: Activating the PDP context");
			PDPSession session = new PDPSession(setAPN, setUser, setPass); // SIM card's APN, user and password
			if (session.getState() != PDPSession.ACTIVATED) {
				session.activate();
			}
		} catch (IOException e) {
			System.out.println("[Main]: Exception during NET configuration phase " + e.toString());
		}

		try {
			MqttClient sampleClient = new MqttClient(broker, clientId, new MemoryPersistence());
			MqttConnectOptions connOpts = new MqttConnectOptions();

			connOpts.setCleanSession(true);
			connOpts.setKeepAliveInterval(120);
			connOpts.setConnectionTimeout(60);
			connOpts.setAutomaticReconnect(true);   // ensure to use refactored PAHO library package naming

			if (brokerUser != null) {connOpts.setUserName(brokerUser);}
			if (brokerPass != null) {connOpts.setPassword(brokerPass.toCharArray());}

			System.out.println("");
			System.out.println("[Main]: Starting incoming message Callback Listener");
			sampleClient.setCallback(new MyMqttCallback());     // listens for incoming MQTT messages

			System.out.println("[Main]: Connecting to broker: " + broker);
			sampleClient.connect(connOpts);
			System.out.println("[Main]: Connected");

			System.out.println("[Main]: Subscribing to topic: '" + pubTopic + "'");
			sampleClient.subscribe(pubTopic, pubQos);           // subscribe to Topic for MQTT messages

			System.out.println("[Main]: Publish to topic: '" + pubTopic + "'");
			System.out.println("[Main]: Publishing message: '" + pubContent + "'");
			MqttMessage message = new MqttMessage(pubContent.getBytes());
			message.setRetained(false);              // you snooze; you lose (the message)
			message.setQos(pubQos);
			sampleClient.publish(pubTopic, message);

			if (message.isRetained()) {
				System.out.println("[Main]: Message will be retained by server.");
			} else {
				System.out.println("[Main]: Message will not be retained by server.");
			}
			System.out.println("[Main]: Message published.\n");

			try {Thread.sleep(2000);} catch (Exception ex) { /* wait 2 seconds before next message... */ }

			sendLocationAsJSON(sampleClient, pubTopic, message, location);

			try {Thread.sleep(2000);} catch (Exception ex) { /* wait 2 seconds before next message... */ }

			sendModulePCBTempReadingAsJSON(sampleClient, pubTopic, message, atc);

			System.out.println("");

			sampleClient.disconnect();
			System.out.println("[Main]: Disconnected from MQTT broker.");
			System.out.println("[Main]: Finished.");
		} catch (MqttException me) {
			System.out.println("[Main]: MqttException system error: ");
			System.out.println("        reason " + me.getReasonCode());
			System.out.println("        msg    " + me.getMessage());
			System.out.println("        loc    " + me.getLocalizedMessage());
			System.out.println("        cause  " + me.getCause());
			System.out.println("        excep  " + me);
			me.printStackTrace();
			me.getCause().printStackTrace();
		}
	}

	private void unlockSim(ATConnection atc) throws IOException {
		sendCommand(atc, "AT+CPIN=" + setPIN + "\r");
		String response = readResponse(atc);
		if (!response.contains("OK")) {
			throw new IOException("AT+CPIN command failed: " + response);
		}
	}

	private boolean attachNetwork(ATConnection atc) throws IOException {
		sendCommand(atc, "AT+CGATT=1\r");
		String response = readResponse(atc);
		return response.contains("OK");
	}

	private void configurePDP(ATConnection atc) throws IOException {
		sendCommand(atc, "AT+CGDCONT=1,\"IP\",\"" + setAPN + "\"\r");
		String response = readResponse(atc);
		if (!response.contains("OK")) {
			throw new IOException("AT+CGDCONT=1 command failed: " + response);
		}
	}

	private void sendCommand(ATConnection atc, String command) throws IOException {
		try {
			// Make sure there is at least a 100ms gap between two AT commands.
			Thread.sleep(100);
		} catch (InterruptedException e) {
			// do nothing
		}

		atc.sendCommand(command);
	}

	private void sendLocationAsJSON(MqttClient sampleClientIn, String pubTopicIn, MqttMessage messageIn,
			String location) {
		try {
			if (location != null) {
				System.out.println("");
				System.out.println("[Main]: Publishing message: '" + location + "'\n");
				messageIn.setPayload((location).getBytes());
				sampleClientIn.publish(pubTopicIn, messageIn);
			}
		} catch (Exception e) {
			System.out.println("[Main]: Exception in sendLocationAsJSON() " + e.toString());
		}
	}

	private void sendModulePCBTempReadingAsJSON(MqttClient sampleClientIn, String pubTopicIn, MqttMessage messageIn,
			ATConnection atc) {
		final boolean DEBUG = false;
		String response = null;
		String payload = null;
		String module_Temperature_degC = null;

		try {
			if (atc != null) {
				atc.sendCommand("AT^SCTM=1,1\r"); // request that we see TEMP in next AT Command
				readResponse(atc);
				atc.sendCommand("AT^SCTM?\r");
				response = readResponse(atc);
			} else {
				System.out.println("[Main]: Warning - could not grab atc() resource to send AT Commands (Temp).");
			}

			if ((response != null) && (response.indexOf("^SCTM: 1,0,") > -1) && (response.indexOf("OK") > -1)) {
				if (DEBUG) {
					System.out.println("[Main]: AT^SCTM - " + response);
				}

				try {
					module_Temperature_degC = response
							.substring(response.indexOf("^SCTM: 1,0,") + "^SCTM: 1,0,".length(),
									response.indexOf("OK") - 2)
							.trim();
				} catch (Exception e) {
					System.out.println("[Main]: Exception in response.substring() parsing " + e.toString() + " '"
							+ response + "'");
				}
			} else {
				System.out.println("[Main]: Warning - error sending AT Commands AT^SCTM=1,1 / AT^SCTM?");
			}

			if (module_Temperature_degC != null) {
				if (DEBUG) {
					System.out.println("[Main]: module_Temperature_degC=" + module_Temperature_degC + "oC");
				}

				/*
				 * JSON - Proper format for a collection is { "key": "value" } see
				 * http://https://jsonlint.com/ { "PCB temp degC": "37" }
				 */
				payload = "{ \"PCB temp degC\": \"" + module_Temperature_degC + "\" }";

				System.out.println("");
				System.out.println("[Main]: Publishing message: '" + payload + "'");
				messageIn.setPayload((payload).getBytes());
				sampleClientIn.publish(pubTopicIn, messageIn);
			} else {
				System.out.println("[Main]: Warning - error parsing response from AT^SCTM?");
			}
		} catch (Exception e) {
			System.out.println("[Main]: Exception in sendModulePCBTempReadingAsJSON() " + e.toString());
		}
	}

	// Attempts to read fresh IMEI value
	public static String readIMEI(ATConnection atc, String clientIdIn) {
		final boolean DEBUG = false;
		String retVal = clientIdIn; // If we can't get new value then reuse old value
		String response = null;

		try {
			if (atc != null) {
				atc.sendCommand("AT+CGSN\r");
				response = readResponse(atc);
			} else {
				System.out.println("[Main]: Warning - could not grab atc() resource to send AT Commands (IMEI).");
			}

			if ((response != null) && (response.indexOf("ERROR") == -1) && (response.indexOf("+CIEV") == -1)
					&& (response.indexOf("^") == -1) && (response.indexOf("\n") > -1)
					&& (response.indexOf("OK") > -1)) {
				if (DEBUG) {
					System.out.println("[Main]: IMEI - " + response);
				}

				try {
					response = response.substring(response.indexOf('\n'), response.indexOf("OK") - 2).trim();
					if (DEBUG) {
						System.out.println("[Main]: IMEI parsed - '" + response + "'");
					}
					Long.parseLong(response); // using this to ensure we just have a number
					retVal = response; // we won't get here if the parseLong above fails
				} catch (Exception e) {
					System.out.println("[Main]: Exception parsing in readIMEI() " + e.toString());
				}
			}

			if (DEBUG) {
				System.out.println("[Main]: retVal (IMEI) = '" + retVal + "'");
			}
		} catch (Exception e) {
			System.out.println("[Main]: Exception in readIMEI() " + e.toString());
		}

		return retVal;
	}

	private static void clearAtBuffer(ATConnection atc) throws IOException {
		byte[] buffer = new byte[1024];
		int read;
		do {
			read = atc.read(buffer, 0, buffer.length);
		} while (read > 0);
	}

	private static String readResponse(ATConnection atc) throws IOException {
		StringBuilder sb = new StringBuilder();
		byte[] buffer = new byte[100];
		int read;
		String response;
		do {
			read = atc.read(buffer, 0, 100);
			response = new String(buffer, 0, read);
			sb.append(response);
		} while (!response.contains("\r\nOK\r\n") && !response.contains("ERROR"));
		return sb.toString();
	}

	private String floatToString(float f) {
		StringBuilder stringBuilder = new StringBuilder();
		if (f < 0) {
			stringBuilder.append('-');
			f = -f;
		}
		stringBuilder.append(Float.valueOf(f).intValue());
		stringBuilder.append('.');
		stringBuilder.append(Float.valueOf(f * 10).intValue() % 10);
		stringBuilder.append(Float.valueOf(f * 100).intValue() % 10);
		return stringBuilder.toString();
	}

	// read configuration properties from .jad file
	private void readSettingsFromJADFile() {
		broker = DemoProperties.getProperty(DemoProperties.BROKER_URL);
		brokerCert = DemoProperties.getProperty(DemoProperties.BROKER_CERTIFICATE);
		brokerUser = DemoProperties.getProperty(DemoProperties.BROKER_USERNAME);
		brokerPass = DemoProperties.getProperty(DemoProperties.BROKER_PASSWORD);
		setPIN = DemoProperties.getProperty(DemoProperties.SIM_PIN);
		setAPN = DemoProperties.getProperty(DemoProperties.SIM_APN);
		setUser = DemoProperties.getProperty(DemoProperties.SIM_USERNAME);
		setPass = DemoProperties.getProperty(DemoProperties.SIM_PASSWORD);
		pubTopic = DemoProperties.getProperty(DemoProperties.MQTT_TOPIC);
		skipGnss = "true".equals(DemoProperties.getProperty(DemoProperties.SKIP_GNSS));

		if ((broker != null) || (brokerCert != null) || (brokerUser != null) || (brokerPass != null) ||
				(setPIN != null) || (setAPN != null) || (setUser != null) || (setPass != null) || (pubTopic != null)) {
			System.out.println("[Main]:  -------- Settings from demo.properties.list --------");
			System.out.println("[Main]:  broker.url:         " + broker);
			System.out.println("[Main]:  broker.certificate: " + brokerCert);
			System.out.println("[Main]:  broker.username:    " + brokerUser);
			System.out.println("[Main]:  broker.password:    " + brokerPass);
			System.out.println("[Main]:  sim.pin:            " + setPIN);
			System.out.println("[Main]:  sim.apn:            " + setAPN);
			System.out.println("[Main]:  sim.username:       " + setUser);
			System.out.println("[Main]:  sim.password:       " + setPass);
			System.out.println("[Main]:  mqtt.topic:         " + pubTopic);
			System.out.println("[Main]:  skip.gnss:          " + skipGnss);
			System.out.println("[Main]:  ----------------------- end ------------------------");
		}
	}

	public static void main(String[] args) {
		new Main().startApp();
	}
}

/* -------------------------------------------------------------------------- */

class MyMqttCallback implements MqttCallback {
	/**
	 * @see MqttCallback#connectionLost(Throwable)
	 */
	@Override
	public void connectionLost(Throwable cause) {
		System.out.println("[MyMqttCallback]: Connection lost because: " + cause);
	}

	/**
	 * @see MqttCallback#deliveryComplete(IMqttDeliveryToken)
	 */
	@Override
	public void deliveryComplete(IMqttDeliveryToken token) {
	}

	/**
	 * @see MqttCallback#messageArrived(String, MqttMessage)
	 */
	@Override
	public void messageArrived(String topic, MqttMessage messageIn) {
		String incomingMessageText;

		try {
			if ((topic != null) && (messageIn != null)) {
				incomingMessageText = new String(messageIn.getPayload());
				System.out.println("[MyMqttCallback]: Message arrived: '" + incomingMessageText + "'");
				System.out.println("[MyMqttCallback]: Subscribe Topic: '" + topic + "'");
			}
		} catch (Exception e) {
			// ignore
		}
	}
}
