/* @start_prolog@
 * Version: %Z% %W% %I% %E% %U%
 * ============================================================================
 *   <copyright
 *   notice="oco-source"
 *   pids="5724-H72,"
 *   years="2010,2012"
 *   crc="2914692611" >
 *   IBM Confidential
 *
 *   OCO Source Materials
 *
 *   5724-H72,
 *
 *   (C) Copyright IBM Corp. 2010, 2012
 *   Copyright 2020 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 *   The source code for the program is not published
 *   or otherwise divested of its trade secrets,
 *   irrespective of what has been deposited with the
 *   U.S. Copyright Office.
 *   </copyright>
 * ============================================================================
 * @end_prolog@
 */
package com.cwm.net;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.security.GeneralSecurityException;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;

import org.eclipse.paho110.client.mqttv311.MqttException;
import org.eclipse.paho110.client.mqttv311.internal.NetworkModule;
import org.eclipse.paho110.client.mqttv311.logging.Logger;
import org.eclipse.paho110.client.mqttv311.logging.LoggerFactory;

import com.cwm.util.GPRSConnectOptions;

import ej.net.util.ssl.SslContextBuilder;

/**
 * A network module for connecting over SSL from Java ME (CLDC profile).
 */
public class SSLMicroNetworkModule implements NetworkModule {

	private static final String DEFAULT_LIST = "/certificates/paho.certificates.list";
	private static final String CERTIFICATES_LIST = "org.eclipse.paho.certificates";
	private static String CERTIFICATE_PATH;

	private String uri;
	private final String host;
	private final int port;
	private SSLSocket connection;
	private InputStream in;
	private OutputStream out;
	private final GPRSConnectOptions connOptions;
	static final String className = SSLMicroNetworkModule.class.getName();
	Logger log = LoggerFactory.getLogger(LoggerFactory.MQTT_CLIENT_MSG_CAT, className);

	/**
	 * Constructs a new SSLMicroNetworkModule using the specified host and port.
	 *
	 * @param host
	 *            the host name to connect to
	 * @param port
	 *            the port to connect to
	 * @param secure
	 *            whether or not to use SSL.
	 */
	public SSLMicroNetworkModule(String host, int port) {
		connOptions = GPRSConnectOptions.getInstance();

		if (connOptions.getConnParamsSocket() != null) {
			this.uri = "ssl://" + host + ":" + port + ";" + connOptions.getConnParamsSocket();
		} else {
			this.uri = "ssl://" + host + ":" + port
					+ ";bearer_type=" + connOptions.getBearerType()
					+ ";access_point=" + connOptions.getAPN()
					+ ";username=" + connOptions.getUser()
					+ ";password=" + connOptions.getPasswd()
					+ ";timeout=" + connOptions.getTimeout();
		}
		this.host = host;
		this.port = port;
	}

	/**
	 * Starts the module, by creating a TCP socket to the server.
	 */
	@Override
	public void start() throws IOException, MqttException {
		final String methodName = "start";
		try {
			log.fine(className, methodName, "252", new Object[] { uri });
			connection = (SSLSocket) getSslContext().getSocketFactory().createSocket();
			connection.connect(new InetSocketAddress(host, port), connOptions.getTimeout());
			connection.setSoLinger(true, 5);

			in = connection.getInputStream();
			out = connection.getOutputStream();
		} catch (IOException | GeneralSecurityException ex) {
			// @TRACE 250=Failed to create TCP socket
			log.fine(className, methodName, "250", null, ex);
			throw new MqttException(
					MqttException.REASON_CODE_SERVER_CONNECT_ERROR, ex);
		}
	}

	@Override
	public InputStream getInputStream() {
		return in;
	}

	@Override
	public OutputStream getOutputStream() {
		return out;
	}

	/**
	 * Stops the module, by closing the TCP socket.
	 */
	@Override
	public void stop() throws IOException {
		in.close();
		out.close();
		connection.close();
	}

	@Override
	public String getServerURI() {              // SJM - added for Paho project version 1.1.0
		String retVal;

		if (this.uri != null) {
			retVal = this.uri;
		} else {
			retVal = "";
		}

		return  retVal;
	}

	private SSLContext getSslContext() throws GeneralSecurityException, IOException {
		SslContextBuilder builder = new SslContextBuilder();
		if (CERTIFICATE_PATH != null) {
			try (InputStream inputStream = new FileInputStream(CERTIFICATE_PATH)) {
				builder.addServerCertificate(inputStream);
			}
		} else {
			String certificatesList = System.getProperty(CERTIFICATES_LIST, DEFAULT_LIST);
			try (InputStream inputStream = SSLMicroNetworkModule.class.getResourceAsStream(certificatesList)) {
				if (inputStream != null) {
					BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));
					for (String certificate = reader.readLine(); certificate != null; certificate = reader.readLine()) {
						if (!certificate.startsWith("#") && !certificate.isEmpty()) {
							builder.addServerCertificate(certificate);
						}
					}
				}
			}
		}
		return builder.build();
	}

	public static void setCertificatePath(String path) {
		CERTIFICATE_PATH = path;
	}

}
