/**
 * Java
 *
 * Copyright 2011-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.ecom.mobile;

import java.io.IOException;

import ej.bon.Immortals;
import ej.ecom.mobile.natives.NativeAT;
import ej.sni.SNI;

/**
 * This class represent an AT connection. It is used to send or receive AT data from virtual UART. A connection session
 * has to be used in the same thread.
 */
public class ATConnection {

	private static ATConnection instance;

	/**
	 * Length of data received buffer.
	 */
	public static final int DATARECEIVED_BUFFER_LENGTH = 1024;

	/**
	 * Data received buffer.
	 */
	public static byte[] DataReceived;

	static {
		DataReceived = (byte[]) Immortals.deepImmortal(new byte[DATARECEIVED_BUFFER_LENGTH]);
	}

	/**
	 * Constructs an AT connection.
	 */
	private ATConnection() {
		// Nothing to do here.
	}

	/**
	 * @return the ATConnection instance.
	 */
	public static synchronized ATConnection getInstance() {
		if (instance == null) {
			instance = new ATConnection();
		}
		return instance;
	}

	/**
	 * Send AT command from the connection.
	 *
	 * @param command
	 *            The AT command as a String to send.
	 * @throws IOException
	 *             If an I/O error occurred or connection closed.
	 */
	public void sendCommand(String command) throws IOException {
		byte[] data = new byte[command.length() + 1];
		SNI.toCString(command, data);
		int res = NativeAT.sendData(data, data.length);
		if (res < 0) {
			throw new IOException(); // an error occurred
		}
	}

	/**
	 * Send AT command from the connection.
	 *
	 * @param data
	 *            The byte array containing the AT command.
	 * @param offset
	 *            The offset of the AT command in the byte array.
	 * @param length
	 *            The length of the byte array.
	 *
	 * @throws IOException
	 *             If an I/O error occurred or connection closed.
	 */
	public void sendCommand(byte[] data, int offset, int length) throws IOException {
		byte[] buffer = new byte[length];
		System.arraycopy(data, offset, buffer, 0, length);
		int res = NativeAT.sendData(buffer, buffer.length);
		if (res < 0) {
			throw new IOException(); // an error occurred
		}
	}

	/**
	 *
	 * Reads up to len bytes of data from the connection into an array of bytes. An attempt is made to read as many as
	 * len bytes, but a smaller number may be read, at most one. The number of bytes actually read is returned as an
	 * integer. This method blocks until input data is available or an exception is thrown. If data is null, a
	 * NullPointerException is thrown. If offset is negative, or length is negative, or offset+length is greater than
	 * the length of the array data, then an IndexOutOfBoundsException is thrown. If length is zero, then no bytes are
	 * read and 0 is returned; otherwise, there is an attempt to read at least one byte. If no byte is available because
	 * the stream is at end of file, the value -1 is returned; otherwise, at least one byte is read and stored into
	 * data. The first byte read is stored into element data[offset], the next one into data[offset+1], and so on. The
	 * number of bytes read is, at most, equal to length. Let k be the number of bytes actually read; these bytes will
	 * be stored in elements data[offset] through data[offset+k-1], leaving elements data[offset+k] through
	 * data[offset+length-1] unaffected. In every case, elements data[0] through data[offset] and elements
	 * data[offset+length] through data[data.length-1] are unaffected. If the first byte cannot be read for any reason
	 * other than end of file, then an IOException is thrown. In particular, an IOException is thrown if the connection
	 * has been closed.
	 *
	 * @param data
	 *            The destination buffer.
	 * @param offset
	 *            The offset of the destination buffer.
	 * @param length
	 *            The maximum length of data to read.
	 * @return Number of data read.
	 *
	 * @throws IOException
	 *             If I/O error occurred or connection closed.
	 * @throws NullPointerException
	 *             If data byte array is null.
	 * @throws IndexOutOfBoundsException
	 *             If offset is negative, or length is negative, or offset+length is greater than the length of the
	 *             array data.
	 */
	public int read(byte[] data, int offset, int length) throws IOException {
		if (data == null) {
			throw new NullPointerException();
		}

		if ((offset < 0) || (length < 0) || (offset + length) > data.length) {
			throw new IndexOutOfBoundsException();
		}

		if (data.length == 0) {
			return 0;
		}

		NativeAT.waitForData();
		int res = 0;
		synchronized (DataReceived) {
			int len = Math.min(length, DataReceived.length);
			res = NativeAT.readData(DataReceived, len);
			if (res > 0) {
				System.arraycopy(DataReceived, 0, data, offset, res);
			} else if (res < 0) {
				throw new IOException();
			}
		}
		return res;
	}

}
