/**
 * Java
 *
 * Copyright 2012-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.ecom.mobile;

import ej.bon.Immortals;
import ej.ecom.mobile.natives.NativePDP;
import ej.sni.SNI;

/**
 * This class represent a PDP session. It's used to activate data mode.
 */
public class PDPSession {

	/**
	 * Deactivated state value
	 */
	public static final int DEACTIVATED = 0;
	/**
	 * Activated state value
	 */
	public static final int ACTIVATED = 1;

	private static final int UNKNOWN_SESSION_ID = -1;

	private final String apn;
	private final String login;
	private final String password;

	/**
	 * Current state of the sessions, either {@link #DEACTIVATED} or {@link #ACTIVATED}.
	 */
	protected int state;

	/**
	 * The ID of the PDP session.
	 */
	protected int id;

	/**
	 * Data buffer for the APN.
	 */
	public static byte[] apnBuffer;

	/**
	 * Data buffer for the login.
	 */
	public static byte[] loginBuffer;

	/**
	 * Data buffer for the secret.
	 */
	public static byte[] pwdBuffer;

	static {
		Immortals.run(new Runnable() {
			@Override
			public void run() {
				apnBuffer = new byte[NativePDP.getAPNMaxSize()];
				loginBuffer = new byte[NativePDP.getLoginMaxSize()];
				pwdBuffer = new byte[NativePDP.getPwdMaxSize()];
			}
		});
	}

	/**
	 * Constructs a PDP session. Default state is {@link #DEACTIVATED}.
	 *
	 * @param apn
	 *            Access Point Name
	 * @param login
	 *            Access Point Name login
	 * @param password
	 *            Access Point Name password
	 */
	public PDPSession(String apn, String login, String password) {
		this.apn = apn;
		this.login = login;
		this.password = password;
		this.id = UNKNOWN_SESSION_ID;
		this.state = DEACTIVATED;
	}

	/**
	 * Activates PDP session. Throws a {@link PDPSessionException} if an error occurred during activation. This method
	 * is thread safe.
	 *
	 * @throws PDPSessionException
	 *             If an error occurred during PDP session activation.
	 */
	public synchronized void activate() throws PDPSessionException {
		if (this.state == ACTIVATED) {
			throw new PDPSessionException("PDP session already activated"); //$NON-NLS-1$
		}

		String apn = this.apn;
		int apnLength = apn.length();
		SNI.toCString(apn, apnBuffer);
		String login = this.login;
		int loginLength = login.length();
		SNI.toCString(login, loginBuffer);
		String pwd = this.password;
		int pwdLength = this.password.length();
		SNI.toCString(pwd, pwdBuffer);
		int res = NativePDP.activate(apnBuffer, apnLength, loginBuffer, loginLength, pwdBuffer, pwdLength);
		if (res < 0) {
			throw new PDPSessionException("Error occurred during PDP session activation"); //$NON-NLS-1$
		} else {
			this.id = res;
			this.state = ACTIVATED;
		}
	}

	/**
	 * Deactivates PDP session. Does nothing if PDP session is already deactivated. Throws a {@link PDPSessionException}
	 * if an error occurred during deactivation. This method is thread safe.
	 *
	 * @throws PDPSessionException
	 *             If an error occurred during PDP session deactivation.
	 */
	public synchronized void deactivate() throws PDPSessionException {
		if (this.state == DEACTIVATED) {
			return;
		}

		int res = NativePDP.deactivate(this.id);
		if (res < 0) {
			throw new PDPSessionException("Error occurred during PDP session deactivation"); //$NON-NLS-1$
		} else {
			this.state = DEACTIVATED;
		}
	}

	/**
	 * @return the current state of the PDP session ({@link #ACTIVATED} or {@link #DEACTIVATED}). This method is thread
	 *         safe.
	 */
	public synchronized int getState() {
		return this.state;
	}

}
