/*
 * Java
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.gnss;

import java.io.IOException;

import ej.sni.NativeIOException;

/**
 * Implementation of the GnssManager function that calls the native function interface.
 */
public class GnssManager {
	private static volatile GnssManager GnssManagerInstance = new GnssManager();
	/**
	 * GnssManager state.
	 */
	private GnssManagerState state;
	private float latitude;
	private float longitude;
	private float altitude;
	private float velocity;
	private long GnssEpochTime;

	/**
	 * enum state machine for GnssManager.
	 */
	public enum GnssManagerState {
		/**
		 * Idle state.
		 */
		IDLE,
		/**
		 * Initialized state.
		 */
		INITIALIZED,
		/**
		 * Positioning polling started.
		 */
		STARTED,
		/**
		 * Positioning polling stopped.
		 */
		STOPPED,
		/**
		 * Close the GnssManager. Necessary if some closure needs to be done in C-side of if you want to stop the
		 * GnssManager and not the application
		 */
		CLOSED;
	}

	static {
		GnssMicroEJNative.GNSS_initialize();
		GnssManager.getInstance().state = GnssManagerState.INITIALIZED;
	}

	/**
	 * Initialize the manager instance.
	 *
	 */
	private GnssManager() {
		this.state = GnssManager.GnssManagerState.IDLE;
		this.GnssEpochTime = 0;
	}

	/**
	 * Read the position. If no data can be read throw a native exception Set the data to invalid if no signal is found
	 * See @fetchPosition.
	 *
	 * @throws IOException
	 *             if no signal was found or reading data failed
	 *
	 */
	public void readPosition() throws IOException {
		assert (this.state == GnssManagerState.STARTED);
		try {
			fetchPosition();
		} catch (NativeIOException e) {
			// We may still update timing data if location is invalid
			if (e.getErrorCode() == GnssMicroEJNative.GNSS_READ_INVALID_DATA_ERR_CODE) {
				updateTime();
			}
			throw e;
		}
		updateTime();
		updateDatas();
	}

	/**
	 * SwitchOn Positioning polling on the board machine.
	 *
	 * * @throws IOException if turning the polling ON on board failed.
	 */
	public void switchOn() throws IOException {
		if (this.state == GnssManagerState.STARTED) {
			return;
		}
		assert (this.state == GnssManagerState.INITIALIZED);

		try {
			GnssMicroEJNative.GNSS_start();
			this.state = GnssManagerState.STARTED;
		} catch (NativeIOException e) {
			throw e;
		}
	}

	/**
	 * SwitchOff Positioning polling on the state machine.
	 *
	 * @throws IOException
	 *             if turning the polling OFF on board failed
	 */
	public void switchOff() throws IOException {
		if (this.state == GnssManagerState.STOPPED) {
			return;
		}
		assert (this.state == GnssManagerState.STARTED);

		try {
			GnssMicroEJNative.GNSS_stop();
			this.state = GnssManagerState.STOPPED;
		} catch (NativeIOException e) {
			throw e;
		}
	}

	/**
	 * Close the manager instance. After this method is called no GnssManager functionality will be available until you
	 * restart the VM.
	 *
	 * @throws IOException
	 *             if closing the GnssManager and its low level component on board failed
	 */
	public void closeManager() throws IOException {
		if (this.state == GnssManagerState.CLOSED) {
			return;
		}
		assert (this.state == GnssManagerState.STOPPED);

		try {
			GnssMicroEJNative.GNSS_close();
			this.state = GnssManagerState.CLOSED;
		} catch (NativeIOException e) {
			throw e;
		}
	}

	/**
	 * Get the latest latitude data in degree.
	 *
	 * @return latitude positioning data.
	 */
	public float getLatitude() {
		return GnssManagerInstance.latitude;
	}

	/**
	 * Return the latest obtained longitude data in degree.
	 *
	 * @return longitude positioning data.
	 */
	public float getLongitude() {
		return GnssManagerInstance.longitude;
	}

	/**
	 * Return the latest obtained altitude in meters data.
	 *
	 * @return longitude positioning data.
	 */
	public float getAltitude() {
		return GnssManagerInstance.altitude;
	}

	/**
	 * Return the latest obtained velocity in m/s data.
	 *
	 * @return velocity positioning data.
	 */
	public float getVelocity() {
		return GnssManagerInstance.velocity;
	}

	/**
	 * Get the epoch time UTC.
	 *
	 * @return Calendar DateTime.
	 */
	public long getTime() {
		return this.GnssEpochTime;
	}

	/**
	 * Return the singleton instance of GnssManager.
	 *
	 * @return singleton instance of GnssManager.
	 */
	public static GnssManager getInstance() {
		return GnssManagerInstance;
	}

	/**
	 * Fetch the latest position received from the GNSS. This function will suspend the current Java thread until a new
	 * position arrive.
	 *
	 * @param None
	 */
	private void fetchPosition() throws IOException {
		GnssMicroEJNative.GNSS_fetchData();
	}

	/**
	 * Write the native timing data received to the LocalizationData attributes.
	 *
	 */
	private void updateTime() {
		this.GnssEpochTime = GnssMicroEJNative.GNSS_getTime();
	}

	/**
	 * Write the native position data received to the LocalizationData attributes If data is invalid set the
	 * CoordinateType to INVALID and data to 0.0.
	 *
	 * @param isValid
	 *            boolean indicating if the localization data is valid or not
	 */
	private void updateDatas() {
		this.latitude = GnssMicroEJNative.GNSS_getLatitude();
		this.longitude = GnssMicroEJNative.GNSS_getLongitude();
		this.altitude = GnssMicroEJNative.GNSS_getAltitude();
		this.velocity = GnssMicroEJNative.GNSS_getVelocity();
	}
}
