/*
 * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2018-2019 IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.security;

import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;

import javax.crypto.JceSecurity;

/**
 * Key factories are used to convert <I>keys</I> (opaque
 * cryptographic keys of type {@code Key}) into <I>key specifications</I>
 * (transparent representations of the underlying key material), and vice
 * versa.
 *
 * <P> Key factories are bi-directional. That is, they allow you to build an
 * opaque key object from a given key specification (key material), or to
 * retrieve the underlying key material of a key object in a suitable format.
 *
 * <P> Multiple compatible key specifications may exist for the same key.
 * For example, a DSA public key may be specified using
 * {@code DSAPublicKeySpec} or
 * {@code X509EncodedKeySpec}. A key factory can be used to translate
 * between compatible key specifications.
 *
 * <P> The following is an example of how to use a key factory in order to
 * instantiate a DSA public key from its encoding.
 * Assume Alice has received a digital signature from Bob.
 * Bob also sent her his public key (in encoded format) to verify
 * his signature. Alice then performs the following actions:
 *
 * <pre>
 * X509EncodedKeySpec bobPubKeySpec = new X509EncodedKeySpec(bobEncodedPubKey);
 * KeyFactory keyFactory = KeyFactory.getInstance("DSA");
 * PublicKey bobPubKey = keyFactory.generatePublic(bobPubKeySpec);
 * Signature sig = Signature.getInstance("DSA");
 * sig.initVerify(bobPubKey);
 * sig.update(data);
 * sig.verify(signature);
 * </pre>
 *
 * <p> Every implementation of the Java platform is required to support the
 * following standard {@code KeyFactory} algorithms:
 * <ul>
 * <li>{@code DiffieHellman}</li>
 * <li>{@code DSA}</li>
 * <li>{@code RSA}</li>
 * </ul>
 * These algorithms are described in the <a href=
 * "{@docRoot}/../technotes/guides/security/StandardNames.html#KeyFactory">
 * KeyFactory section</a> of the
 * Java Cryptography Architecture Standard Algorithm Name Documentation.
 * Consult the release documentation for your implementation to see if any
 * other algorithms are supported.
 *
 * @author Jan Luehe
 *
 * @see Key
 * @see PublicKey
 * @see PrivateKey
 * @see java.security.spec.KeySpec
 * @see java.security.spec.DSAPublicKeySpec
 * @see java.security.spec.X509EncodedKeySpec
 *
 * @since 1.2
 */

public class KeyFactory {

	// The algorithm associated with this key factory
	private final String algorithm;

	// The provider implementation (delegate)
	private volatile KeyFactorySpi spi;

	private KeyFactory(String algorithm) throws NoSuchAlgorithmException {
		this.algorithm = algorithm;
		this.spi = JceSecurity.PROVIDER.getKeyFactorySpi(algorithm);
	}

	/**
	 * Returns a KeyFactory object that converts public/private keys of the
	 * specified algorithm.
	 *
	 * @param algorithm
	 *            the name of the requested key algorithm. See the KeyFactory
	 *            section in the <a href=
	 *            "{@docRoot}/../technotes/guides/security/StandardNames.html#KeyFactory">
	 *            Java Cryptography Architecture Standard Algorithm Name
	 *            Documentation</a> for information about standard algorithm names.
	 *
	 * @return the new KeyFactory object.
	 *
	 * @exception NoSuchAlgorithmException
	 *                if the KeyFactorySpi has no implementation for the specified
	 *                algorithm.
	 *
	 * @see Provider
	 */
	public static KeyFactory getInstance(String algorithm)
			throws NoSuchAlgorithmException {
		return new KeyFactory(algorithm);
	}

	/**
	 * Gets the name of the algorithm
	 * associated with this {@code KeyFactory}.
	 *
	 * @return the name of the algorithm associated with this
	 * {@code KeyFactory}
	 */
	public final String getAlgorithm() {
		return this.algorithm;
	}

	/**
	 * Generates a private key object from the provided key specification
	 * (key material).
	 *
	 * @param keySpec the specification (key material) of the private key.
	 *
	 * @return the private key.
	 *
	 * @exception InvalidKeySpecException if the given key specification
	 * is inappropriate for this key factory to produce a private key.
	 */
	public final PrivateKey generatePrivate(KeySpec keySpec)
			throws InvalidKeySpecException {
		return spi.engineGeneratePrivate(keySpec);
	}

}
