/*
 * Copyright (c) 1999, 2012, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package javax.net.ssl;

import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;

public class KeyManagerFactory {

	// The provider implementation (delegate)
	private final KeyManagerFactorySpi factorySpi;

	// The name of the key management algorithm.
	private final String algorithm;

	public final static String getDefaultAlgorithm() {
		String type;
		type = System.getProperty("ssl.KeyManagerFactory.algorithm");
		if (type == null) {
			type = "SunX509";
		}
		return type;
	}

	private KeyManagerFactory(KeyManagerFactorySpi factorySpi, String algorithm) {
		this.factorySpi = factorySpi;
		this.algorithm = algorithm;
	}

	public final String getAlgorithm() {
		return this.algorithm;
	}

	public static final KeyManagerFactory getInstance(String algorithm)
			throws NoSuchAlgorithmException {
		String spiImpl = System.getProperty("KeyManagerFactory." + algorithm.toUpperCase());
		try {
			KeyManagerFactorySpi keyManagerFactorySpi = (KeyManagerFactorySpi) Class.forName(spiImpl).newInstance();
			return new KeyManagerFactory(keyManagerFactorySpi, algorithm);
		} catch (Exception e) {
			throw new NoSuchAlgorithmException(algorithm + " KeyManagerFactory not available", e);
		}
	}


	public final void init(KeyStore ks, char[] password) throws
	KeyStoreException, NoSuchAlgorithmException,
	UnrecoverableKeyException {
		factorySpi.engineInit(ks, password);
	}


	public final KeyManager[] getKeyManagers() {
		return factorySpi.engineGetKeyManagers();
	}
}
