/*
 * Copyright (c) 1999, 2012, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package javax.net.ssl;

import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

public class SSLContext {

	private final SSLContextSpi contextSpi;

	private final String protocol;

	/**
	 * Creates an SSLContext object.
	 *
	 * @param contextSpi the delegate
	 * @param protocol the protocol
	 */
	protected SSLContext(SSLContextSpi contextSpi, String protocol) {
		this.contextSpi = contextSpi;
		this.protocol = protocol;
	}

	private static SSLContext defaultContext;

	public static synchronized SSLContext getDefault()
			throws NoSuchAlgorithmException {
		if (defaultContext == null) {
			defaultContext = SSLContext.getInstance("Default");
		}
		return defaultContext;
	}

	public static SSLContext getInstance(String protocol)
			throws NoSuchAlgorithmException {
		String impl = System.getProperty("SSLContext."+protocol.toUpperCase());
		try{
			SSLContextSpi contextSpi = (SSLContextSpi)Class.forName(impl).newInstance();
			return new SSLContext(contextSpi, protocol);
		}catch(Exception e){
			throw new NoSuchAlgorithmException(protocol + " SSLContext not available", e);
		}
	}

	public final String getProtocol() {
		return this.protocol;
	}

	public final void init(KeyManager[] km, TrustManager[] tm,
			SecureRandom random)
					throws KeyManagementException {
		contextSpi.engineInit(km, tm, random);
	}

	public final SSLSocketFactory getSocketFactory() {
		return contextSpi.engineGetSocketFactory();
	}

	public final SSLServerSocketFactory getServerSocketFactory() {
		return contextSpi.engineGetServerSocketFactory();
	}
}
