/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

/**
 * @file
 * @brief LLFS implementation over QAPI.
 */

/* Includes ------------------------------------------------------------------*/

#include <stdint.h>
#include <string.h>
#include "qapi_fs.h"
#include "LLFS_impl.h"
#include "LLFS_File_impl.h"
#include "microej_async_worker.h"
#include "fs_helper.h"
#include "fs_configuration.h"
#include "microej_allocator.h"

#ifdef __cplusplus
	extern "C" {
#endif

/**
 * Sanity check between the expected version of the configuration and the actual version of
 * the configuration.
 * If an error is raised here, it means that a new version of the CCO has been installed and
 * the configuration fs_configuration.h must be updated based on the one provided
 * by the new CCO version.
 */
#if FS_CONFIGURATION_VERSION != 1

	#error "Version of the configuration file fs_configuration.h is not compatible with this implementation."

#endif

#ifdef LLFS_DEBUG
#define LLFS_DEBUG_LOG_INFO(...) gina_uwlog_printf("[LLFS][INFO]: " __VA_ARGS__);
#define LLFS_DEBUG_LOG_ERROR(...) gina_uwlog_printf("[LLFS][ERROR]: " __VA_ARGS__);
#else
#define LLFS_DEBUG_LOG_INFO(...) ((void) 0)
#define LLFS_DEBUG_LOG_ERROR(...) ((void) 0)
#endif // LLFS_DEBUG

#define LLFS_NORMAL_PERMISSIONS (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)

static uint8_t not_supported_message[] = "Not supported";

/**
 * @brief Simple time conversion from seconds to LLFS_date_t.
 * 
 * @param time the time to convert in seconds.
 * @return LLFS_date_t the date converted from time.
 */
static LLFS_date_t time_to_date(uint32_t time) {
	LLFS_date_t result = {0};
	result.second = time % 60;
	time /= 60;
	result.minute = time % 60;
	time /= 60;
	result.hour = time % 24;
	time /= 24;
	result.day = time % 30;
	time /= 30;
	result.month = time % 12;
	time /= 12;
	result.year = time;
	return result;
}

/**
 * Set the size of the file referenced by the given file descriptor into size_out.
 * Returns LLFS_NOK on error or LLFS_OK on success.
 */
static int FS_size_of_file(int fd, uint64_t* size_out) {
	jint fs_err;
	int res;
	struct qapi_FS_Stat_Type_s buffer;

	qapi_FS_Status_t status = qapi_FS_Stat_With_Handle(fd, &buffer);

	if(QAPI_OK != status){
		*size_out = 0x0ll;
		res = LLFS_NOK;
	} else {
		*size_out = (uint64_t)buffer.st_size;
		res = LLFS_OK;
	}

	return res;
}

void LLFS_IMPL_get_last_modified_action(MICROEJ_ASYNC_WORKER_job_t* job) {
	FS_last_modified_t* params = (FS_last_modified_t*)job->params;
	uint8_t* path = params->path;
	LLFS_date_t* out_date = &params->date;

	struct qapi_FS_Stat_Type_s buffer;
	LLFS_date_t date = {0};
	params->result = LLFS_NOK; // error by default

	qapi_FS_Status_t status = qapi_FS_Stat(path, &buffer);

	if (QAPI_OK == status) {
		date = time_to_date(buffer.st_mtime);
		out_date->second = date.second;
		out_date->minute = date.minute;
		out_date->hour = date.hour;
		out_date->day = date.day;
		out_date->month = date.month;
		out_date->year = date.year;
		params->result = LLFS_OK;
	}
}

void LLFS_IMPL_set_read_only_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	
	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] readonly set on %s (status %d)", __FILE__, __LINE__, params->path, params->result);
}

void LLFS_IMPL_create_action(MICROEJ_ASYNC_WORKER_job_t* job) {
	FS_create_t* params = (FS_create_t*)job->params;
	uint8_t* path = params->path;
	int fd = 0;

	params->result = LLFS_NOK; // error by default

	/*
	 * Open file with:
	 * - force created: O_CREAT
	 * - fail if directory: O_WRONLY
	 * - fail if the file exists: O_EXCL
	 */
	qapi_FS_Status_t status = qapi_FS_Open_With_Mode(path, O_CREAT | O_WRONLY | O_EXCL, LLFS_NORMAL_PERMISSIONS, &fd);

	/* test return function */
	if(QAPI_OK != status) {
		if(QAPI_OK == qapi_FS_Close(fd)){
			params->result = LLFS_OK;  // success
		} else {
			params->error_code = status;
			params->error_message = "Creation failed";
		}
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] : create file %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_open_directory_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] open dir %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_read_directory_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_read_directory_t* params = (FS_read_directory_t*)job->params;
	int32_t directory_ID = params->directory_ID;
	uint8_t* path = params->path;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] read dir result %s (status %d)", __FILE__, __LINE__, path, params->result);
}


void LLFS_IMPL_close_directory_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_close_directory_t* params = (FS_close_directory_t*)job->params;
	int32_t directory_ID = params->directory_ID;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] close dir (status %d)", __FILE__, __LINE__, params->result);
}

void LLFS_IMPL_rename_to_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_rename_to_t* params = (FS_rename_to_t*)job->params;
	uint8_t* path = params->path;
	uint8_t* new_path = params->new_path;

    qapi_FS_Status_t status = qapi_FS_Rename(path, new_path);

	if (QAPI_OK == status){
		params->result = LLFS_OK;
	}
	else {
		params->result = LLFS_NOK;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] rename : old name %s, new name %s (status %d)", __FILE__, __LINE__, path, new_path, params->result);
}

void LLFS_IMPL_get_length_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path64_operation_t* params = (FS_path64_operation_t*)job->params;
	uint8_t* path = params->path;

	params->result = LLFS_NOK; // error by default

	struct qapi_FS_Stat_Type_s buffer;
	qapi_FS_Status_t status = qapi_FS_Stat(path, &buffer);
	if(QAPI_OK == status){
		params->result = buffer.st_size;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] length of %s : %ld", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_exist_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	struct qapi_FS_Stat_Type_s buffer;
	qapi_FS_Status_t status = qapi_FS_Stat(path, &buffer);

	if (QAPI_OK == status){
		params->result = LLFS_OK;
	}
	else {
		params->result = LLFS_NOK;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] exist file %s : %d, %d",  __FILE__, __LINE__, path, params->result, status);
}

void LLFS_IMPL_get_space_size_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_get_space_size_t* params = (FS_get_space_size_t*)job->params;
	uint8_t* path = params->path;
	int32_t space_type = params->space_type;

	params->result = LLFS_NOK; //error by default

    struct qapi_FS_Statvfs_Type_s buffer;
    qapi_FS_Status_t status = qapi_FS_Statvfs(path, &buffer);

	if(QAPI_OK == status){
		/* f_blocks, f_bfree and f_bavail are defined in terms of f_bsize */
		jlong scale_factor = (jlong)buffer.f_bsize;

		switch (space_type) {
			case LLFS_FREE_SPACE:
				params->result = ((unsigned long)buffer.f_bfree * scale_factor);
				break;
			case LLFS_TOTAL_SPACE:
				params->result = ((unsigned long)buffer.f_blocks * scale_factor);
				break;
			case LLFS_USABLE_SPACE:
				params->result = ((unsigned long)buffer.f_bavail * scale_factor);
				break;
		}
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] : get space %d size on %s : %ld", __FILE__, __LINE__, space_type, path, params->result);
}

void LLFS_IMPL_make_directory_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

    qapi_FS_Status_t status = qapi_FS_Mk_Dir(path, (S_IRWXU | S_IRWXG | S_IRWXO));
	if (QAPI_OK == status){
		params->result = LLFS_OK;
	}
	else{
		params->result = LLFS_NOK;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] : mkdir %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_is_hidden_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	// Hidden files are not supported.
	params->result = LLFS_NOK;

	LLFS_DEBUG_LOG_INFO("[%s:%u] : is hidden %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_is_directory_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] : is directory %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_is_file_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] : is file %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_set_last_modified_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_last_modified_t* params = (FS_last_modified_t*)job->params;
	uint8_t* path = params->path;

	// Error: not supported
	params->result = LLFS_NOK;
	SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("LLFS_DEBUG [%s:%u] : set last modified date on %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_delete_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_path_operation_t* params = (FS_path_operation_t*)job->params;
	uint8_t* path = params->path;

	qapi_FS_Status_t status_unlink = qapi_FS_Unlink(path);
	qapi_FS_Status_t status_rmdir = qapi_FS_Rm_Dir(path);

	if((QAPI_OK == status_unlink) || (QAPI_OK == status_rmdir)){
		params->result = LLFS_OK;
	}
	else {
		params->result = LLFS_NOK;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] : delete %s (status %d)", __FILE__, __LINE__, path, params->result);
}

void LLFS_IMPL_is_accessible_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_is_accessible_t* params = (FS_is_accessible_t*)job->params;
	uint8_t* path = params->path;

    // Error: not supported
	params->result = LLFS_NOK;
    SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] : is accessible %s access %d (status %d)", __FILE__, __LINE__, path, 0, params->result);
}

void LLFS_IMPL_set_permission_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_set_permission_t* params = (FS_set_permission_t*)job->params;
	uint8_t* path = params->path;

    // Error: not supported
	params->result = LLFS_NOK;
    SNI_throwNativeException(0, (const char*) &not_supported_message);

	LLFS_DEBUG_LOG_ERROR("[%s:%u] : set permission %s access %d enable %d owner %d (status %d)", __FILE__, __LINE__, path, 0, 0, 0, params->result);
}

void LLFS_File_IMPL_open_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_open_t* params = (FS_open_t*)job->params;
	uint8_t* path = params->path;
	uint8_t mode = params->mode;
    qapi_FS_Status_t status;

	int open_flags;
	int fd;

	microej_malloc(8); // Workaround CC0231G-82

	params->result = LLFS_NOK; // error by default
	params->error_code = LLFS_NOK;
	params->error_message = NULL;

	switch (mode) {
		case LLFS_FILE_MODE_READ:
			open_flags = O_RDONLY;
			break;

		case LLFS_FILE_MODE_WRITE:
			open_flags = O_CREAT| O_WRONLY| O_TRUNC ;
			break;

		case LLFS_FILE_MODE_APPEND:
			open_flags = O_CREAT | O_WRONLY | O_APPEND;
			break;

		default:
			params->error_code = mode;
			params->error_message = "Invalid opening mode";
			return;
	}

	/* open file according to internal mode and internal permission*/
    status = qapi_FS_Open_With_Mode(path, open_flags, LLFS_NORMAL_PERMISSIONS, &fd);

	/* test return function */
	if ((fd == -1) || (QAPI_OK != status)){
		params->error_code = status;
		params->error_message = "Open failed";
	}
	else {
		params->result = fd;  // no error
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] : open %s mode %d (status %d, %d)", __FILE__, __LINE__, path, mode, params->result, status);
}

void LLFS_File_IMPL_write_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_write_read_t* params = (FS_write_read_t*)job->params;
	int32_t fd = params->file_id;
	uint8_t* data = params->data;
	int32_t length = params->length;
	uint32_t written_count = 0;

	qapi_FS_Status_t status = qapi_FS_Write(fd, data, length, &written_count);

	// - status != QAPI_OK when an error is detected
	// - written_count == 0 when nothing was written (for instance if the FS is full);
	//     in this case, let's check that we indeed wanted to write more than 0 bytes
	//     to determine if something bad happened
	if ((QAPI_OK != status) || (written_count == 0 && length > 0)){
		params->result = LLFS_NOK; // error
		params->error_code = status;
		params->error_message = "Write failed";
	}
	else {
		params->result = written_count;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] write file content %d - %d bytes to write (status %d)\n",  __FILE__, __LINE__, fd, length, params->result);
}

void LLFS_File_IMPL_read_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_write_read_t* params = (FS_write_read_t*)job->params;
	int32_t fd = params->file_id;
	uint8_t* data = params->data;
	int32_t length = params->length;
	uint32_t read_count = 0;

	qapi_FS_Status_t status = qapi_FS_Read(fd, data, length, &read_count);

	if (QAPI_OK != status){
		params->result = LLFS_NOK; // error
		params->error_code = status;
		params->error_message = "Read failed";
	}
	else if (read_count == 0){
		params->result = LLFS_EOF;// EOF
	}
	else {
		params->result = read_count;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] read file content %d - %d bytes to read (status %d)",  __FILE__, __LINE__, fd, length, params->result);
}

void LLFS_File_IMPL_close_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_close_t* params = (FS_close_t*)job->params;
	int32_t fd = params->file_id;

	qapi_FS_Status_t status = qapi_FS_Close(fd);

	if(QAPI_OK != status){
		params->result = LLFS_NOK;
		params->error_code = status;
		params->error_message = "Close failed";
	}
	else {
		params->result = LLFS_OK;
	}

	LLFS_DEBUG_LOG_INFO("[%s:%u] close file %d (status %d errno %d)\n",  __FILE__, __LINE__, fd, params->result);
}

void LLFS_File_IMPL_skip_action(MICROEJ_ASYNC_WORKER_job_t* job){
	FS_skip_t* params = (FS_skip_t*)job->params;
	int32_t fd = params->file_id;
	int64_t n = params->n;
	qapi_FS_Status_t status = 0;

	// Basically do a lseek with the given offset.
	// Before, checks the value of the offset:
	//	- don't go out of the bounds of the file:
	//		- compare the new position (current+offset) with the size of the file
	// 	- can be cast on the type defined by lseek

	int fs_err;
	uint64_t file_size;

	// Get the size of the file
	fs_err = FS_size_of_file(fd, &file_size);
	if(fs_err == LLFS_OK){
		// No error when computing the size of the file.

		//Get current file position.
		off_t actual_offset = 0;
		status = qapi_FS_Seek(fd, 0, SEEK_CUR, &actual_offset);
		if(((off_t)-1 != actual_offset) && (QAPI_OK == status)){
			long current_position = actual_offset;
			//Saturate the offset to limit the new position to the size of the file.
			uint64_t max_offset = file_size - current_position;
			if(n > max_offset){
				n = max_offset;
			}

			// Convert given offset in a type accepted by lseek
			off_t offset = (off_t)n;

			// Check if the conversion from long long int to off_t is correct
			if(offset < 0 || offset != n){
				// An overflow occurs, saturate the value
				offset = INT32_MAX;
			}

			// Skip the bytes !
			qapi_FS_Status_t status = qapi_FS_Seek(fd, offset, SEEK_CUR, &actual_offset);
			if(((off_t) -1 != actual_offset) && (QAPI_OK == status)){
				// Skip done
				off_t new_position = actual_offset;
				params->skipped_count = new_position - current_position;
				params->result = LLFS_OK;

				LLFS_DEBUG_LOG_INFO("[%s:%u] skip %ld bytes on %d (status %d skip count %ld)",  __FILE__, __LINE__, n, fd, params->result, params->skipped_count);

				return;
			}
		}
	}

	// Error occurred
	params->skipped_count = 0;
	params->result = LLFS_NOK;
	params->error_code = qapi_FS_Last_Error();
	params->error_message = "Skip failed";

	LLFS_DEBUG_LOG_ERROR("[%s:%u] error skip %ld bytes on %d (status %d)",  __FILE__, __LINE__, n, fd, params->result);
}

void LLFS_File_IMPL_available_action(MICROEJ_ASYNC_WORKER_job_t* job) {
	FS_available_t* params = (FS_available_t*)job->params;
	int32_t fd = params->file_id;

	params->result = LLFS_NOK; // error by default
	params->error_code = LLFS_NOK;
	params->error_message = NULL;

	int fs_err;
	uint64_t file_size;

	// Get file size
	fs_err = FS_size_of_file(fd, &file_size);
	if(fs_err == 0) {
		// Get current position
		off_t offset = 0;
		qapi_FS_Status_t status = qapi_FS_Seek(fd, 0, SEEK_CUR, &offset);
		if(((off_t)-1 != offset) && (QAPI_OK == status)) {
			long current_position = offset;
			uint64_t available = file_size - current_position;
			if((int)available != available) {
				//overflow when casting on int: return max value
				params->result = INT32_MAX;
			} else {
				params->result = (int)available;
			}
		} else {
			// error during ftell: just update fs_err for debug/verbose.
			fs_err = (int)status;
		}
	}
	// else: error when computing the size of the file.

	LLFS_DEBUG_LOG_INFO("[%s:%u] available %d bytes on %d",  __FILE__, __LINE__, params->result, fd);
}

void LLFS_File_IMPL_flush_action(MICROEJ_ASYNC_WORKER_job_t* job) {
	FS_available_t* params = (FS_available_t*)job->params;
	int32_t fd = params->file_id;

	params->result = LLFS_OK; // nothing to do

	LLFS_DEBUG_LOG_INFO("[%s:%u] flush %d (status=%d)",  __FILE__, __LINE__, fd, params->result);
}

#ifdef __cplusplus
	}
#endif
