/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

/**
 * @file
 * @brief Platform implementation specific macro.
 * @author MicroEJ Developer Team
 * @version 1.1.1
 * @date 7 February 2020
 */

#ifndef  LLNET_CONFIGURATION_H
#define  LLNET_CONFIGURATION_H

#include <stdint.h>
#include <time.h>
#include "qapi_socket.h"

// #error "This header must be customized with platform specific configuration. Remove this #error when done. This file is not modified when a new version of the CCO is installed."

/**
 * @brief Compatibility sanity check value.
 * This define value is checked in the implementation to validate that the version of this configuration
 * is compatible with the implementation.
 *
 * This value must not be changed by the user of the CCO.
 * This value must be incremented by the implementor of the CCO when a configuration define is added, deleted or modified.
 */
#define LLNET_CONFIGURATION_VERSION (2)

/**
 * By default all the llnet_* functions are mapped on the BSD functions.
 * Documentation (signature, etc.) for each llnet_* function is available
 * in the mapped BSD function documentation.
 */
#define llnet_accept		qapi_accept
#define llnet_bind			qapi_bind
#define llnet_close			qapi_socketclose
#define llnet_connect		qapi_connect
#define llnet_fcntl			qapi_fcntl
#define llnet_gethostname	qapi_gethostname
#define llnet_getpeername	qapi_getpeername
#define llnet_getsockname	qapi_getsockname
#define llnet_getsockopt	qapi_getsockopt
#define llnet_htonl			_htonl
#define llnet_htons			_htons
#define llnet_ioctl			qapi_ioctl
#define llnet_listen		qapi_listen
#define llnet_ntohs			_ntohs
#define llnet_recv 			qapi_recv
#define llnet_recvfrom		qapi_recvfrom
#define llnet_send			qapi_send
#define llnet_sendto		qapi_sendto
#define llnet_setsockopt	qapi_setsockopt
#define llnet_socket		qapi_socket
#define llnet_shutdown		qapi_shutdown
#define llnet_select        qapi_select

#define IPPROTO_TCP 0

/**
 * Don't modify the LLNET_AF_* constants.
 */
#define LLNET_AF_IPV4	(0x1)
#define LLNET_AF_IPV6	(0x2)
#define LLNET_AF_DUAL	((LLNET_AF_IPV4) | (LLNET_AF_IPV6))

/**
 *  Address family:
 *  - LLNET_AF_IPV4 for only IPv4
 *  - LLNET_AF_IPV6 for only IPv6
 *  - LLNET_AF_DUAL for both IPv4 and IPv6
 */
#define LLNET_AF (LLNET_AF_IPV4)

/**
 * Returns the errno value for the given file descriptor.
 * Given file descriptor may be -1 if no file descriptor is defined.
 */
#define llnet_errno     qapi_errno

/*
 * Initialize the network stack.
 * Returns 0 on success, -1 on error.
 * This method may block and must be called in another RTOS task.
 * See async_select task.
 * By default this macro does nothing.
 */
#define llnet_init()	(0)

/*
 * Returns true (bool) if we can call the services of the network stack without jeopardizing the system,
 * otherwise returns false (bool).
 * Returning true does not mean that a network interface is up.
 * By default this macro always return true.
 */
#define llnet_is_ready()	(true)

/** @brief Set this define if the system sends SIGPIPE signal when a connection is closed by the remote host. */
#if defined(__linux__) || defined(__QNXNTO__)
#define LLNET_IGNORE_SIGPIPE
#endif

#define INADDR_LOOPBACK 0

/**
 * Enable network debug trace
 */
// #define LLNET_DEBUG

/**
 * If ioctl() cannot be used to implement the LLNET_STREAMSOCKETCHANNEL_IMPL_available() function,
 * define USE_MSG_PEEK_FOR_AVAILABLE and NET_EMBEDDED_AVAILABLE_BUFFER_SIZE.
 */
#define USE_MSG_PEEK_FOR_AVAILABLE
#define NET_EMBEDDED_AVAILABLE_BUFFER_SIZE (4096)

/**
 * Define USE_IOCTL_FOR_BLOCKING_OPTION to use ioctl() instead of fcntl() to configure socket
 * blocking or non-blocking mode.
 */
//#define USE_IOCTL_FOR_BLOCKING_OPTION


#if LLNET_AF & LLNET_AF_IPV6
/**
 * Only one IPV6 interface is supported
 */
#define LLNET_IPV6_INTERFACE_NAME "eth0"

#endif

/*** Definitions for Qualcomm API port ***/

/**
 * @brief This macro converts an unsigned short integer from host byte order to network byte order.
 * 
 */
#define _htons(x) ((uint16_t)((((uint16_t)(x) & 0x00ff) << 8) | (((uint16_t)(x) & 0xff00) >> 8)))

/**
 * @brief This macro converts an unsigned short integer from network byte order to host byte order.
 * 
 */
#define _ntohs(x) ((uint16_t)((((uint16_t)(x) & 0x00ff) << 8) | (((uint16_t)(x) & 0xff00) >> 8)))

/**
 * @brief This macro converts a long integer from host byte order to network byte order.
 * 
 */
#define _htonl(x) ((uint32_t)((((uint32_t)(x) & 0x000000ff) << 24) | (((uint32_t)(x) & 0x0000ff00) << 8) | (((uint32_t)(x) & 0x00ff0000) >> 8) | (((uint32_t)(x) & 0xff000000) >> 24)))

/**
 * @brief Use the getsockopt() function to set blocking option.
 * 
 */
#define USE_GETSOCKOPT_FOR_BLOCKING_OPTION

/**
 * @brief Define the address for binding to all local interfaces.
 * 
 */
#define INADDR_ANY ((in_addr_t) 0x00000000)

/**
 * @brief Define the in_addr_t type.
 * 
 */
typedef uint32_t in_addr_t;

typedef unsigned socklen_t;

struct ip_mreq
{
	struct in_addr imr_multiaddr;
	struct in_addr imr_interface;
};

#endif // LLNET_CONFIGURATION_H
