/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 */

#include <LLNET_DNS_impl.h>
#include <string.h>
#include <sni.h>
#include <microej_allocator.h>
#include <LLNET_DNS_exs82.h>
#include "qapi_dss.h"
#include "qapi_status.h"
#include "qapi_socket.h"
#include "qapi_dnsc.h"
#include "LLNET_CONSTANTS.h"
#include "LLNET_ERRORS.h"
#include "LLNET_Common.h"

#ifdef __cplusplus
extern "C" {
#endif

#define DNS_CLIENT_TIMEOUT (10000)
#define DNS_NUM_ADDRS (5)

static uint8_t dns_client_started = 0;

static int32_t start_dns_client(void) {
	qapi_Status_t ret = 0;

	LLNET_DEBUG_TRACE("%s", __func__);

	ret = qapi_Net_DNSc_Command(QAPI_NET_DNS_START_E);
	if (QAPI_DSS_SUCCESS != ret) {
		LLNET_DEBUG_TRACE("%s Failed to start DNS client, returned %d\n", __func__, ret);
		return -1;
	}

	dns_client_started = 1;
	return 0;
}

int32_t LLNET_DNS_addDnsServer(qapi_DSS_Hndl_t netctrl_dss_handle) {
	qapi_Status_t ret = 0;
	char dns_address[16] = { 0x00 };
	qapi_DSS_Addr_Info_t info_ptr[DNS_NUM_ADDRS];

	LLNET_DEBUG_TRACE("%s", __func__);

	// Start DNS client if not already up
	if (!dns_client_started) {
		if(0 != start_dns_client()) {
			return J_EUNKNOWN;
		}
	}

	ret = qapi_DSS_Get_IP_Addr(netctrl_dss_handle, info_ptr, DNS_NUM_ADDRS);
	if (QAPI_OK != ret)	{
		LLNET_DEBUG_TRACE("%s: Get IP address failed with status = %d", __func__, ret);
		return -1;
	}

	if (!info_ptr[0].iface_addr_s.valid_addr) {
		LLNET_DEBUG_TRACE("%s: Address is not valid", __func__);
		return -1;
	}

	inet_ntop(AF_INET,&info_ptr[0].dnsp_addr_s.addr,dns_address,sizeof(dns_address));

	ret = qapi_Net_DNSc_Add_Server(&dns_address, QAPI_NET_DNS_ANY_SERVER_ID);
	if (QAPI_OK != ret)	{
		LLNET_DEBUG_TRACE("%s: Add DNS server failed with status = %d", __func__, ret);
		return -1;
	}

	return 0;
}

int32_t LLNET_DNS_IMPL_getHostByAddr(int8_t* host, int32_t offset, int32_t length, uint8_t retry) {
	int32_t i_return = 0;
	int32_t java_thread;
	char* pc_host = microej_malloc(length + 1);
	struct ip46addr* ipaddr = microej_malloc(sizeof(struct ip46addr));

	if ((NULL == host) || (offset >= length)) {
		return J_EINVAL;
	}

	// Start DNS client if not already up
	if (!dns_client_started) {
		if(0 != start_dns_client()) {
			return J_EUNKNOWN;
		}
	}

	strlcpy(pc_host, (const char*) host, length + 1);

	LLNET_DEBUG_TRACE("%s (%s length=%d retry=%hhu)\n", __func__, pc_host, length, retry);

	ipaddr->type = AF_INET;
	int32_t err;

	qapi_Net_DNSc_Set_Client_timeout(DNS_CLIENT_TIMEOUT);

	/* Try to resolve the host name */
	err = qapi_Net_DNSc_Reshost(pc_host, ipaddr);
	LLNET_DEBUG_TRACE("%s (qapi_Net_DNSc_Reshost returned %d)\n", __func__, err);
	if (err == 0) {
		/* retrieve the host name resolution */
		memcpy(host, &ipaddr->a.addr4, length - offset);

		LLNET_DEBUG_TRACE("%s (%hhu.%hhu.%hhu.%hhu length=%d retry=%hhu)\n", __func__, host[0], host[1], host[2],
				host[3], length, retry);

		/* success */
		i_return = (int32_t) strlen(host);
	} else {
		/* an error occurred while resolving the host name */
		i_return = (int32_t) J_EHOSTUNKNOWN;
	}

	LLNET_DEBUG_TRACE("%s (i_return=%d)\n", __func__, i_return);

	return (i_return);
}

int32_t LLNET_DNS_IMPL_getHostByNameAt(int32_t index, int8_t* host, int32_t offset, int32_t length, uint8_t retry) {
	int32_t i_return;

	if ((NULL == host) || (offset >= length)) {
		return J_EINVAL;
	}

	LLNET_DEBUG_TRACE("%s\n", __func__);

	/* there is only one name for one */
	i_return = LLNET_DNS_IMPL_getHostByAddr(host, offset, length, retry);
	if (i_return > 0) {
		// No error occurred
		i_return = 4;
	}

	return i_return;
}

int32_t LLNET_DNS_IMPL_getHostByNameCount(int8_t* hostname, int32_t offset, int32_t length, uint8_t retry) {
	if ((NULL == hostname) || (offset >= length)) {
		return J_EINVAL;
	}

	LLNET_DEBUG_TRACE("%s\n", __func__);
	/* there is only one name for one */
	return (1);
}

#ifdef __cplusplus
}
#endif
