/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

/**
 * @file
 * @brief OS Abstraction Layer ThreadX implementation
 * @author MicroEJ Developer Team
 * @version 1.0.0
 * @date 12 May 2020
 */

#include <stdint.h>
#include "osal.h"
#include "microej_allocator.h"

/**
 * @brief From the ThreadX documentation: specifies the number of 32-bit word(s) for the queue message size. Valid sizes are values between 1 and 16 inclusive.
 * 
 */
#define MESSAGE_SIZE 1

/**
 * @brief Convert time from milliseconds to ThreadX ticks.
 *
 * @param[in] time in milliseconds
 *
 * @return time in ThreadX ticks
 */
static uint32_t OSAL_ThreadX_convert_time_to_tick(uint32_t milliseconds);

/**
 * @brief Create an OS task and start it.
 *
 * @param[in] entry_point function called at task startup
 * @param[in] name the task name
 * @param[in] stack task stack declared using OSAL_task_stack_declare() macro
 * @param[in] priority task priority
 * @param[in] parameters task entry parameters. NULL if no entry parameters
 * @param[in,out] handle pointer on a task handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_task_create(OSAL_task_entry_point_t entry_point, uint8_t* name, OSAL_task_stack_t stack, int32_t priority,
		void* parameters, OSAL_task_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

#ifdef THREADX_MODULES
	if (TX_SUCCESS != txm_module_object_allocate((void*) &handle->tx_thread, sizeof(TX_THREAD))) {
		return OSAL_NOMEM;
	}
#endif // THREADX_MODULES

	if (TX_SUCCESS
			!= tx_thread_create(handle->tx_thread, (CHAR* ) name, (void (*)(ULONG) ) entry_point, (uint32_t ) parameters,
					(void * ) stack.stack, (uint32_t ) stack.size, (uint32_t ) priority, (uint32_t ) priority - 1, TX_NO_TIME_SLICE,
					TX_AUTO_START)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Delete an OS task and start it.
 *
 * @param[in] handle pointer on the task handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_task_delete(OSAL_task_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_thread_delete(handle->tx_thread) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Get the handle of the current OS task.
 *
 * @param[in,out] handle pointer on a task handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_task_get_current(OSAL_task_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	handle->tx_thread = tx_thread_identify();

	return OSAL_OK;
}

/**
 * @brief Create an OS queue with a predefined queue size.
 *
 * @param[in] name queue name
 * @param[in,out] handle pointer on a queue handle
 * @param[in] queue structure containing address and size declared using OSAL_queue_declare() macro
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_queue_create(uint8_t* name, OSAL_queue_handle_t* handle, OSAL_queue_t queue) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

#ifdef THREADX_MODULES
	if (TX_SUCCESS != txm_module_object_allocate(&handle->tx_queue, sizeof(TX_QUEUE))) {
		return OSAL_NOMEM;
	}
#endif // THREADX_MODULES

	if (tx_queue_create(handle->tx_queue, (CHAR *) name, MESSAGE_SIZE, (VOID *) queue.addr,
			(uint32_t)((uint32_t) queue.size * sizeof(void *))) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Delete an OS queue.
 *
 * @param[in] handle pointer on the queue handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_queue_delete(OSAL_queue_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_queue_delete(handle->tx_queue) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Post a message in an OS queue.
 *
 * @param[in] handle pointer on the queue handle
 * @param[in] msg message to post in the message queue
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_queue_post(OSAL_queue_handle_t* handle, void* msg) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_queue_send(handle->tx_queue, &msg, TX_NO_WAIT) != TX_SUCCESS) {
		return OSAL_NOMEM;
	}

	return OSAL_OK;
}

/**
 * @brief Fetch a message from an OS queue. Blocks until a message arrived or a timeout occurred.
 *
 * @param[in] handle pointer on the queue handle
 * @param[in,out] msg message fetched from the OS queue
 * @param[in] timeout maximum time to wait for message arrival, OSAL_INFINITE_TIME for infinite timeout
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_queue_fetch(OSAL_queue_handle_t* handle, void** msg, uint32_t timeout) {
	uint32_t timeout_in_tick = OSAL_ThreadX_convert_time_to_tick(timeout);

	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_queue_receive(handle->tx_queue, msg, timeout_in_tick) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Create an OS counter semaphore with a semaphore count initial value.
 *
 * @param[in] name counter semaphore name
 * @param[in] initial_count counter semaphore initial count value
 * @param[in] max_count counter semaphore maximum count value
 * @param[in,out] handle pointer on a counter semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_counter_semaphore_create(uint8_t* name, uint32_t initial_count, uint32_t max_count,
		OSAL_counter_semaphore_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

#ifdef THREADX_MODULES
	if (TX_SUCCESS != txm_module_object_allocate(&handle->tx_semaphore_counter, sizeof(TX_SEMAPHORE))) {
		return OSAL_NOMEM;
	}
#endif // THREADX_MODULES

	if (tx_semaphore_create(handle->tx_semaphore_counter, (CHAR *) name, (uint32_t) initial_count) != TX_SUCCESS) {
		return OSAL_NOMEM;
	}

	(void) max_count;
	return OSAL_OK;
}

/**
 * @brief Delete an OS counter semaphore.
 *
 * @param[in] handle pointer on the counter semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_counter_semaphore_delete(OSAL_counter_semaphore_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_semaphore_delete(handle->tx_semaphore_counter) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Take operation on OS counter semaphore. Block the current task until counter semaphore
 * become available or timeout occurred. Decrease the counter semaphore count value by 1 and
 * block the current task if count value equals to 0.
 *
 * @param[in] handle pointer on the counter semaphore handle
 * @param[in] timeout maximum time to wait until the counter semaphore become available, OSAL_INFINITE_TIME for infinite timeout
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_counter_semaphore_take(OSAL_counter_semaphore_handle_t* handle, uint32_t timeout) {
	uint32_t timeout_in_tick = OSAL_ThreadX_convert_time_to_tick(timeout);

	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_semaphore_get(handle->tx_semaphore_counter, timeout_in_tick) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Give operation on OS counter semaphore. Increase the counter semaphore count value by 1 and unblock the current task if count value.
 * equals to 0.
 *
 * @param[in] handle pointer on the counter semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_counter_semaphore_give(OSAL_counter_semaphore_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_semaphore_put(handle->tx_semaphore_counter) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Create an OS binary semaphore with a semaphore count initial value (0 or 1).
 *
 * @param[in] name counter semaphore name
 * @param[in] initial_count counter semaphore initial count value
 * @param[in,out] handle pointer on a binary semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_binary_semaphore_create(uint8_t* name, uint32_t initial_count, OSAL_binary_semaphore_handle_t* handle) {
	uint8_t status = 0;

	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

#ifdef THREADX_MODULES
	if (TX_SUCCESS != txm_module_object_allocate((void*) &handle->tx_semaphore, sizeof(TX_SEMAPHORE))) {
		return OSAL_NOMEM;
	}
#endif // THREADX_MODULES

	/* Create the semaphore */
	if (TX_SUCCESS != tx_semaphore_create(handle->tx_semaphore, (CHAR * ) name, (uint32_t ) initial_count)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Delete an OS binary semaphore.
 *
 * @param[in] handle pointer on the binary semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_binary_semaphore_delete(OSAL_binary_semaphore_handle_t* handle) {
	if (NULL == handle->tx_semaphore) {
		return OSAL_WRONG_ARGS;
	}

	if (TX_SUCCESS != tx_semaphore_delete(handle->tx_semaphore)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Take operation on OS binary semaphore. Block the current task until binary semaphore
 * become available or timeout occurred. Decrease the binary semaphore count value by 1 and
 * block the current task if count value equals to 0.
 *
 * @param[in] handle pointer on the binary semaphore handle
 * @param[in] timeout maximum time to wait until the binary semaphore become available, OSAL_INFINITE_TIME for infinite timeout
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_binary_semaphore_take(OSAL_binary_semaphore_handle_t* handle, uint32_t timeout) {
	uint32_t timeout_in_tick = OSAL_ThreadX_convert_time_to_tick(timeout);

	if (NULL == handle->tx_semaphore) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_semaphore_get(handle->tx_semaphore, timeout_in_tick) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Give operation on OS binary semaphore. Increase the binary semaphore count value by 1 and unblock the current task if count value.
 * equals to 0.
 *
 * @param[in] handle pointer on the binary semaphore handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_binary_semaphore_give(OSAL_binary_semaphore_handle_t* handle) {
	if (NULL == handle->tx_semaphore) {
		return OSAL_WRONG_ARGS;
	}

	if (TX_SUCCESS != tx_semaphore_put(handle->tx_semaphore)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Create an OS mutex.
 *
 * @param[in] name mutex name
 * @param[in,out] handle pointer on a mutex handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_mutex_create(uint8_t* name, OSAL_mutex_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

#ifdef THREADX_MODULES
	if (TX_SUCCESS != txm_module_object_allocate((void*) &handle->tx_mutex, sizeof(TX_MUTEX))) {
		return OSAL_NOMEM;
	}
#endif // THREADX_MODULES

	if (TX_SUCCESS != tx_mutex_create(handle->tx_mutex, (CHAR * ) name, TX_INHERIT)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Delete an OS mutex.
 *
 * @param[in] handle pointer on the mutex handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_mutex_delete(OSAL_mutex_handle_t* handle) {
	if (NULL == handle->tx_mutex) {
		return OSAL_WRONG_ARGS;
	}

	if (TX_SUCCESS != tx_mutex_delete(handle->tx_mutex)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Take operation on OS mutex.
 *
 * @param[in] handle pointer on the mutex handle
 * @param[in] timeout maximum time to wait until the mutex become available, OSAL_INFINITE_TIME for infinite timeout
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_mutex_take(OSAL_mutex_handle_t* handle, uint32_t timeout) {
	uint32_t timeout_in_tick = OSAL_ThreadX_convert_time_to_tick(timeout);

	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}
	
	if (TX_SUCCESS != tx_mutex_get(handle->tx_mutex, timeout_in_tick)) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Give operation on OS mutex.
 *
 * @param[in] handle pointer on the mutex handle
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_mutex_give(OSAL_mutex_handle_t* handle) {
	if (NULL == handle) {
		return OSAL_WRONG_ARGS;
	}

	if (tx_mutex_put(handle->tx_mutex) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

/**
 * @brief Disable the OS scheduler context switching. Prevent the OS from
 * scheduling the current thread calling #OSAL_disable_context_switching while
 * the OS scheduling is already disable has an undefined behavior. This method
 * may be called from an interrupt.
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_disable_context_switching(void) {
	return OSAL_NOT_IMPLEMENTED;
}

/**
 * @brief Reenable the OS scheduling that was disabled by #OSAL_disable_context_switching.
 * This method may be called from an interrupt.
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_enable_context_switching(void) {
	return OSAL_NOT_IMPLEMENTED;
}

/**
 * @brief Asleep the current task during specified number of milliseconds.
 *
 * @param[in] milliseconds number of milliseconds
 *
 * @return operation status (@see OSAL_status_t)
 */
OSAL_status_t OSAL_sleep(uint32_t milliseconds) {
	uint32_t timeout_in_tick = OSAL_ThreadX_convert_time_to_tick(milliseconds);

	if (tx_thread_sleep(timeout_in_tick) != TX_SUCCESS) {
		return OSAL_ERROR;
	}

	return OSAL_OK;
}

static uint32_t OSAL_ThreadX_convert_time_to_tick(uint32_t milliseconds) {
	uint32_t time_to_tick_timeout;

	if (milliseconds == OSAL_INFINITE_TIME) {
		time_to_tick_timeout = TX_WAIT_FOREVER;
	} else {
		if (milliseconds == 0) {
			time_to_tick_timeout = 0UL;
		} else {
			time_to_tick_timeout = ((milliseconds * TX_TIMER_TICKS_PER_SECOND) / 1000UL);
			if (time_to_tick_timeout == 0UL) {
				time_to_tick_timeout = 1UL;
			}
		}
	}

	return time_to_tick_timeout;
}
