/*
 * C
 *
 * Copyright 2020 MicroEJ Corp.
 * This software has been designed or modified by MicroEJ Corp.
 * MicroEJ Corp. grants to THALES the non-exclusive right to freely use, modify and distribute this software.
 *
 */

#include "gina.h"
#include "qapi_spi_master.h"
#include "qapi_timer.h"
#include "qapi_tlmm.h"
#include "gpio_map.h"

#define SPI_POOL_SIZE 30720
#define SPI_BUFFER_SIZE 1024
#define MESSAGE_LENGTH 10

static qapi_SPIM_Config_t config;
static qapi_SPIM_Descriptor_t spi_desc[2];

void *spi_handle = NULL;

static TX_BYTE_POOL *spi_byte_pool = NULL;
ULONG Spi_free_mem[SPI_POOL_SIZE];

static uint8_t *tx_buf;
static uint8_t *rx_buf;

static uint8_t check_gpio(uint32_t pin) {
	qapi_Status_t  status = QAPI_OK;

	qapi_GPIO_ID_t gpio_id;
	qapi_TLMM_Config_t tlmm_config;

	tlmm_config.pin = pin;
	tlmm_config.func = 0;
	tlmm_config.dir = QAPI_GPIO_OUTPUT_E;
	tlmm_config.pull= QAPI_GPIO_NO_PULL_E;
	tlmm_config.drive = QAPI_GPIO_2MA_E;
	
	status = qapi_TLMM_Get_Gpio_ID(&tlmm_config, &gpio_id);
	if (QAPI_OK != status) {
		gina_uwlog_printf("Get ID: pin = %d  status = %d", gpio_id, status);
		return 1;
	}
	
	status = qapi_TLMM_Config_Gpio(gpio_id, &tlmm_config);
	if (QAPI_OK != status) {
		gina_uwlog_printf("Config: GPIO = %d  status = %d", tlmm_config.pin, status);
		return 2;
	}

	status = qapi_TLMM_Drive_Gpio(gpio_id, tlmm_config.pin, QAPI_GPIO_LOW_VALUE_E);
	if (QAPI_OK != status) {
		gina_uwlog_printf("Drive: GPIO = %d  status = %d", tlmm_config.pin, status);
		return 4;
	}

	status = qapi_TLMM_Release_Gpio_ID(&tlmm_config, gpio_id);
	if (QAPI_OK != status) {
		gina_uwlog_printf("Release: GPIO = %d  status = %d", tlmm_config.pin, status);
		return 8;
	}

	return 0; // Success
}

static uint8_t check_gpios() {
	uint8_t total_status = 0;

	uint32_t pin_mosi = 12;
	uint32_t pin_miso = 13;
	uint32_t pin_cs = 14;
	uint32_t pin_clk = 15;
	// uint32_t pin_mosi = gpio_mapping_table[USER_GPIO16].nGpioNumber;
	// uint32_t pin_miso = gpio_mapping_table[USER_GPIO17].nGpioNumber;
	// uint32_t pin_cs = gpio_mapping_table[USER_GPIO18].nGpioNumber;
	// uint32_t pin_clk = gpio_mapping_table[USER_GPIO19].nGpioNumber;

	uint32_t gpios[] = { pin_mosi, pin_miso, pin_cs, pin_clk };

	uint8_t status;
	uint32 gpio;

	for (int i = 0; i < 4; i++) {
		gpio = gpios[i];
		status = check_gpio(gpio);
		if (0 != status) {
			gina_uwlog_printf("GPIO %d unavailable (status = %d)", gpio, status);
		}
		total_status |= status;
	}

	return total_status;
}

static void spi_client_callback(uint32_t status, void *context) {
	gina_uwlog_printf("status=%d", status);
	uint8_t *response = (uint8_t *)context;
	for (int i = 0; i < 10; i++)
	{
		gina_uwlog_printf("%d", response[i]);
	}
}

void qapi_spi_master_test(void) {
	uint8_t *read_buffer;
	qapi_SPIM_Config_t SPI_config;
	TX_BYTE_POOL *byte_spi;

	SPI_config.SPIM_Mode = QAPI_SPIM_MODE_0_E;
	SPI_config.SPIM_CS_Polarity = QAPI_SPIM_CS_ACTIVE_LOW_E;
	SPI_config.SPIM_endianness = SPI_BIG_ENDIAN;
	SPI_config.SPIM_Bits_Per_Word = 8;
	SPI_config.SPIM_Slave_Index = 0;
	SPI_config.Clk_Freq_Hz = 10000000;
	SPI_config.CS_Clk_Delay_Cycles = 0;
	SPI_config.Inter_Word_Delay_Cycles = 0;
	SPI_config.SPIM_CS_Mode = QAPI_SPIM_CS_KEEP_ASSERTED_E;
	SPI_config.loopback_Mode = 1;

	txm_module_object_allocate(&spi_byte_pool, sizeof(TX_BYTE_POOL));
	tx_byte_pool_create(spi_byte_pool, "byte_pool", Spi_free_mem, SPI_POOL_SIZE);
	tx_byte_allocate(spi_byte_pool, (VOID **)&tx_buf, MESSAGE_LENGTH, TX_NO_WAIT);
	tx_byte_allocate(spi_byte_pool, (VOID **)&rx_buf, MESSAGE_LENGTH, TX_NO_WAIT);

	for (int i = 0; i < MESSAGE_LENGTH; i++)
	{
		tx_buf[i] = i;
	}

	qapi_Status_t status;
	void *spi_handle;
	status = qapi_SPIM_Open(QAPI_SPIM_INSTANCE_1_E, &spi_handle);
	gina_uwlog_printf("qapi_SPIM_Open status=%d", status);
	status = qapi_SPIM_Power_On(spi_handle);
	gina_uwlog_printf("qapi_SPIM_Power_On status=%d", status);

	spi_desc[0].tx_buf = tx_buf;
	spi_desc[0].rx_buf = rx_buf;
	spi_desc[0].len = MESSAGE_LENGTH;

	status = qapi_SPIM_Full_Duplex(spi_handle, &SPI_config, &spi_desc[0], 1, (qapi_SPIM_Callback_Fn_t)spi_client_callback, (void *)rx_buf, NULL);
	gina_uwlog_printf("qapi_SPIM_Full_Duplex status=%d", status);

	qapi_Timer_Sleep(10, QAPI_TIMER_UNIT_SEC, true);
}

int qapi_spi_master_test_all(void) {
	qapi_Timer_Sleep(500, QAPI_TIMER_UNIT_MSEC, true);
	gina_uwlog_set_level(GINA_UWLOG_DBG);

	uint8_t gpios_free;
	gina_uwlog_printf("Checking GPIOs...");
	gpios_free = check_gpios();
	gina_uwlog_printf("%s", 0 == gpios_free ? "Success" : "Failed");

	if (0 == gpios_free) {
		gina_uwlog_printf("SPI test start");
		check_gpios();
		qapi_spi_master_test();
		gina_uwlog_printf("SPI test end");
	}

	return 0;
}
